class PygameRenderError(Exception):
    pass


class PygameRenderResourceError(PygameRenderError):
    pass


class PygameEventError(PygameRenderError):
    pass


class PygameEventHandlerError(PygameEventError):
    pass



from abc import ABC, abstractmethod, ABCMeta
from dataclasses import dataclass
from typing import Callable, Iterable, Optional, Generator, Self
from os import get_terminal_size

from beautiful_repr import StylizedMixin, Field
from colorama import Style

from sim32.geometry import Vector
from sim32.interfaces import IUpdatable, IRenderRersourceKeeper, IAvatar, IRenderActivatorFactory
from sim32.errors.render_errors import UnsupportedResourceError
from sim32.tools import ReportAnalyzer, BadReportHandler, Report, Arguments, CustomArgumentFactory, get_collection_with_reduced_nesting_level_by


@dataclass
class ResourcePack:
    """
    Dataclass for transport representation of atomic data for rendering in
    positional form.

    Has two attributes: resource - data to render and point - position to render
    the resource.
    """

    resource: any
    point: any


@dataclass
class StylishResourcePack(ResourcePack):
    """ResourcePack dataclass with style annotation."""

    style: any


class IRenderResourceHandler(ABC):
    """Resource pack handler interface in conjunction with surface and render."""

    @abstractmethod
    def __call__(self, resource_pack: ResourcePack, surface: any, render: 'BaseRender') -> None:
        """Resource pack handling method with an environment."""


class RenderResourceHandler(IRenderResourceHandler, ABC):
    """
    Base class that template implements the RenderResourceHandler interface.

    Templates the main handling method by analyzing the processing support of
    a pack and its environment.
    """

    _report_analyzer = ReportAnalyzer(
        (BadReportHandler(UnsupportedResourceError, "Resource Handler can't handle resource"), )
    )

    def __call__(self, resource_pack: ResourcePack, surface: any, render: 'BaseRender') -> None:
        self._report_analyzer(self.is_support_to_handle(resource_pack, surface, render))
        self._handle(resource_pack, surface, render)

    def is_support_to_handle(self, resource_pack: ResourcePack, surface: any, render: 'BaseRender') -> Report:
        """Method for obtaining analysis of handling conditions."""

        return Report(True)

    @abstractmethod
    def _handle(self, resource_pack: ResourcePack, surface: any, render: 'BaseRender') -> None:
        """Unconditional Packet handling method."""


class ResourceHandlerWrapper(RenderResourceHandler, StylizedMixin):
    """RenderResourceHandler delegating handling to another RenderResourceHandler."""

    _repr_fields = (Field('resource_handler'), )

    def __init__(self, resource_handler: IRenderResourceHandler):
        self.resource_handler = resource_handler

    def is_support_to_handle(self, resource_pack: ResourcePack, surface: any, render: 'BaseRender') -> Report:
        return (
            self.resource_handler.is_support_to_handle(resource_pack, surface, render)
            if hasattr(self.resource_handler, 'is_support_to_handle') else Report(True)
        )

    def _handle(self, resource_pack: ResourcePack, surface: any, render: 'BaseRender') -> None:
        self.resource_handler(resource_pack, surface, render)

    @classmethod
    def create_decorator_by(cls, *args, **kwargs) -> Callable[[IRenderResourceHandler], Self]:
        """Decorator creation method to create ResourceHandlerWrapper via @."""

        def decorator(resource_handler: IRenderResourceHandler):
            return cls(resource_handler, *args, **kwargs)

        return decorator


class TypedResourceHandler(ResourceHandlerWrapper):
    """Resource Handler Wrapper using types as handling conditions."""

    _repr_fields = (Field(
        'supported_resource_type',
        value_getter=lambda handler, _: handler.supported_resource_type.__name__
    ), )

    def __init__(self, resource_handler: IRenderResourceHandler, supported_resource_type: type):
        super().__init__(resource_handler)
        self.supported_resource_type = supported_resource_type

    def is_support_to_handle(self, resource_pack: ResourcePack, surface: any, render: 'BaseRender') -> Report:
        return (
            Report(isinstance(resource_pack.resource, self.supported_resource_type)) and
            super().is_support_to_handle(resource_pack, surface, render)
        )


class IRender(ABC):
    """Render interface for rendering data from resource packs."""

    @abstractmethod
    def __call__(self, resource_pack: ResourcePack) -> None:
        """Native method of rendering a single render pack."""

    @abstractmethod
    def draw_resource_pack(self, resource_pack: ResourcePack) -> None:
        """Render pack drawing method."""

    @abstractmethod
    def draw_scene(self, resource_packs: Iterable[ResourcePack]) -> None:
        """Method for rendering an entire scene from resource packs."""

    @abstractmethod
    def clear_surfaces(self) -> None:
        """Method that allows you to clear surfaces from rendered entities."""


class BaseRender(IRender, ABC):
    """Class that abstractly implements the Render interface."""

    @property
    @abstractmethod
    def surfaces(self) -> tuple:
        """Property of surfaces used by the render."""

    def __call__(self, resource_pack: ResourcePack) -> None:
        self.draw_resource_pack(resource_pack)

    def draw_scene(self, resource_packs: Iterable[ResourcePack]) -> None:
        for surface in self.surfaces:
            self._clear_surface(surface)

            for resource_pack in resource_packs:
                self._draw_resource_pack_on(surface, resource_pack)

    def draw_resource_pack(self, resource_pack: ResourcePack) -> None:
        for surface in self.surfaces:
            self._draw_resource_pack_on(surface, resource_pack)

    def clear_surfaces(self) -> None:
        for surface in self.surfaces:
            self._clear_surface(surface)

    @abstractmethod
    def _draw_resource_pack_on(self, surface: any, resource_pack: ResourcePack) -> None:
        """Atomic rendering method resource packs on the surface."""

    @abstractmethod
    def _clear_surface(self, surface: any) -> None:
        """Atomic single surface cleaning method."""


class ResourceHandlingChainMeta(ABCMeta):
    """
    Metaclass for collecting resource handlers from a class at the time of
    creation of this very class.

    Gathers handlers in _resource_handlers attribute.
    """

    _resource_handlers: Optional[tuple]

    def __new__(cls, class_name: str, super_classes: tuple, attributes: dict):
        render_type = super().__new__(cls, class_name, super_classes, attributes)

        render_type._resource_handlers = (
            tuple(render_type.__get_resource_handlers_from(attributes))
            + render_type._get_resource_handlers_of_parents()
        )

        return render_type

    @staticmethod
    def resource_handler(
        *args_for_factory,
        wrapper_factory: Optional[ResourceHandlerWrapper] = None,
        **kwargs_for_factory,
    ) -> Callable[[IRenderResourceHandler], ResourceHandlerWrapper]:
        """Decorator for system adding a resource handler to a class."""

        def decorator(resource_handler: IRenderResourceHandler) -> ResourceHandlerWrapper | Arguments:
            # Arguments here to initialize handler by metaclass
            return (wrapper_factory if wrapper_factory else Arguments.create_via_call)(
                resource_handler,
                *args_for_factory,
                **kwargs_for_factory
            )

        return decorator

    def _get_resource_handlers_of_parents(cls) -> tuple[IRenderResourceHandler]:
        """Method for collecting all available handlers in the inheritance tree."""

        return sum(
            tuple(
                parent_type._resource_handlers for parent_type in cls.__bases__
                if hasattr(parent_type, '_resource_handlers')
            ),
            tuple()
        )

    def __get_resource_handlers_from(cls, attributes: dict) -> Generator[IRenderResourceHandler, any, None]:
        """Method for collecting handlers from attribute space."""

        for attribute_name, attribute_value in attributes.items():
            if isinstance(attribute_value, RenderResourceHandler):
                yield attribute_value
            elif isinstance(attribute_value, Arguments):
                resource_handler = cls._resource_handler_wrapper_factory(
                    *attribute_value.args,
                    **attribute_value.kwargs
                )
                setattr(cls, attribute_name, resource_handler)
                yield resource_handler


resource_handler = ResourceHandlingChainMeta.resource_handler


class Render(BaseRender, ABC, metaclass=ResourceHandlingChainMeta):
    """BaseRender child class that delegates rendering of packs to pack handlers."""

    _resource_handler_wrapper_factory = ResourceHandlerWrapper

    def _draw_resource_pack_on(self, surface: any, resource_pack: ResourcePack) -> None:
        for resource_handler in self._resource_handlers:
            if resource_handler.is_support_to_handle(resource_pack, surface, self):
                resource_handler(resource_pack, surface, self)


class SurfaceKeeper:
    """Stub class for classes inheriting from Render."""

    def __init__(self, surfaces: Iterable):
        self._surfaces = tuple(surfaces)

    @property
    def surfaces(self) -> tuple:
        return self._surfaces


class ConsoleCell:
    """Console cell view class. Stores the immediate content and its style."""

    def __init__(self, sign: str, style: Iterable[str] = tuple()):
        self.sign = sign
        self.style = style

    def __str__(self) -> str:
        return str().join((self.style, self.sign, Style.RESET_ALL if self.style else ''))

    @property
    def sign(self) -> str:
        """Cell content property."""

        return self.__sign

    @sign.setter
    def sign(self, sign: str) -> None:
        self.__sign = sign[:1]

    @property
    def style(self) -> str:
        """
        Cell style property.
        By definition, be an ANSI code or something based on it.
        """

        return self.__style

    @style.setter
    def style(self, style: Iterable[str]) -> None:
        self.__style = str().join(style)


class ConsoleScene:
    """
    Content class of the active console \"frame\".
    Stores data as a two-dimensional missive.
    """

    def __init__(self, size: Iterable[int], default_empty_cell: ConsoleCell):
        self.reset(size, default_empty_cell)

    def __str__(self) -> str:
        return str().join(map(str, get_collection_with_reduced_nesting_level_by(3, self.__cell_table)))

    @property
    def default_empty_cell(self) -> ConsoleCell:
        """Cell property defining all unoccupied."""

        return self.__default_empty_cell

    @default_empty_cell.setter
    def default_empty_cell(self, default_empty_cell: ConsoleCell) -> None:
        for y in range(size[1]):
            for x in range(size[0]):
                if self[(x, y)] is self.__default_empty_cell:
                    self[(x, y)] = default_empty_cell

        self.__default_empty_cell = default_empty_cell

    @property
    def size(self) -> tuple[int, int]:
        """Scene size as 2D array size."""

        return self.__size

    @size.setter
    def size(self, size: Iterable[int]) -> None:
        if size[1] < self.__size[1]:
            self.__cell_table[size[1] - 1:] = tuple()

        if size[0] > self.__size[0]:
            for column in self.__cell_table:
                column.extend((self.default_empty_cell, ) * (size[0] - self.__size[0]))
        elif size[0] < self.__size[0]:
            for column in self.__cell_table:
                column[size[0] - 1:] = tuple()

        if size[1] > self.__size[1]:
            self.__cell_table.extend((self.default_empty_cell, ) * size[0])

        self.__size = tuple(size)

    def __getitem__(self, key: Iterable[int]) -> ConsoleCell:
        return self.__cell_table[key[1]][key[0]]

    def __setitem__(self, key: Iterable[int], value: ConsoleCell) -> None:
        if key[0] < self.size[0] and key[1] < self.size[1]:
            self.__cell_table[key[1]][key[0]] = value

    def reset(self, size: Optional[Iterable[int]] = None, default_empty_cell: Optional[ConsoleCell] = None) -> None:
        """Scene data reset and override method."""

        self.__size = tuple(size) if size is not None else self.__size

        self.__default_empty_cell = (
            default_empty_cell if default_empty_cell is not None
            else self.__default_empty_cell
        )

        self.__cell_table = [
            [self.__default_empty_cell] * self.__size[0]
            for _ in range(self.__size[1])
        ]


class ConsoleRender(IRender):
    """
    Console rendering class.

    In the absence of access to the console by the system or due to some other
    reason, it can still be active, but without any results. Not designed for
    high fps rendering since the console may flash in the first and last line
    due to slow output (relative to the difference in time between two \"frames\")
    to this very console.

    Accepts strings and a ConsoleCell as atomic render data.
    """

    def __init__(
        self,
        empty_cell: ConsoleCell,
        default_filled_cell: ConsoleCell = ConsoleCell('#')
    ):
        self.default_filled_cell = default_filled_cell
        self.__scene = ConsoleScene(self._console_size, empty_cell)

    @property
    def empty_cell(self) -> ConsoleCell:
        """Cell property that defines unoccupied other cells."""

        return self.__scene.default_empty_cell

    @empty_cell.setter
    def empty_cell(self, empty_cell: ConsoleCell) -> None:
        self.__scene.default_empty_cell = empty_cell

    def __call__(self, resource_pack: ResourcePack) -> None:
        self.draw_resource_pack(resource_pack)

    def draw_resource_pack(self, resource_pack: ResourcePack) -> None:
        self._insert_resource_pack_into_scene(resource_pack)
        self._draw_current_scene()

    def draw_scene(self, resource_packs: Iterable[ResourcePack]) -> None:
        self._clear_scene()

        for resource_pack in resource_packs:
            self._insert_resource_pack_into_scene(resource_pack)

        self._draw_current_scene()

    def clear_surfaces(self) -> None:
        self._clear_scene()

    @property
    def _scene(self) -> ConsoleScene:
        """Active render scene property."""

        return self.__scene

    @property
    def _console_size(self) -> tuple[int, int]:
        """Real size property of the console."""

        try:
            terminal_size = get_terminal_size()
            return (terminal_size.columns, terminal_size.lines - 1)
        except OSError:
            return (0, 0)

    def _draw_current_scene(self) -> None:
        """Method for outputting a saved frame to the console."""

        if all(self.__scene.size):
            print('\n' + str(self.__scene))

    def _clear_scene(self) -> None:
        """Method for clearing the active frame from all elements lying on it."""

        self.__scene.reset(self._console_size)

    def _insert_resource_pack_into_scene(self, resource_pack: ResourcePack) -> None:
        """Method for inserting data from the resource pack into the active frame."""

        resource_pack = self._get_usable_resource_pack(resource_pack)

        self.__scene[resource_pack.point] = str(resource_pack.resource)

    def _get_usable_resource_pack(self, resource_pack: ResourcePack) -> ResourcePack:
        """Method for getting the given resource pack into the correct form for rendering."""

        return ResourcePack(
            resource=self._get_usable_cell_from(resource_pack),
            point=tuple(map(
                int,
                (
                    resource_pack.point.coordinates[:2]
                    if isinstance(resource_pack.point, Vector) else resource_pack.point
                )
            ))
        )

    def _get_usable_cell_from(self, resource_pack: ResourcePack) -> ConsoleCell:
        """
        Atomic method for converting data from a resource pack into the correct
        form of ConsoleCell.
        """

        sign = None
        style = None

        if isinstance(resource_pack.resource, ConsoleCell):
            sign = resource_pack.resource.sign
            if resource_pack.resource.style:
                style = resource_pack.resource.style
        else:
            if hasattr(resource_pack, 'style') and isinstance(resource_pack.style, str) and resource_pack.style:
                style = resource_pack.style

            sign = (
                resource_pack.resource
                if isinstance(resource_pack.resource, str) and resource_pack.resource
                else self.default_filled_cell.sign
            )

        if not style:
            style = self.default_filled_cell.style

        return ConsoleCell(sign, style)


class RenderActivator(IUpdatable):
    """
    Unit class that activates the rendering of scenes from resource packs.

    Gets packages from input render_resource_keepers and delegates rendering to
    input renders.
    """

    def __init__(self, render_resource_keepers: Iterable[IRenderRersourceKeeper], renders: Iterable[Render]):
        self.render_resource_keepers = render_resource_keepers
        self.renders = tuple(renders)

    def update(self) -> None:
        for render in self.renders:
            render.draw_scene(self._get_render_resource_packs())

    def _get_render_resource_packs(self) -> Generator[ResourcePack, any, None]:
        """Method to get all resource packs from stored render resource keepers."""

        for keeper in self.render_resource_keepers:
            yield from keeper.render_resource_packs


class CustomRenderActivatorFactory(CustomArgumentFactory, IRenderActivatorFactory):
    """Class of factory render activators."""

    def __call__(
        self,
        render_resource_keepers: Iterable[IRenderRersourceKeeper],
        redners: Iterable[Render],
        *args,
        **kwargs
    ) -> RenderActivator:
        return super().__call__(render_resource_keepers, redners, *args, **kwargs)



from dataclasses import dataclass
from typing import Iterable

from sim32.geometry import Vector
from sim32.tools import RGBAColor


@dataclass
class ColorRenderResource:
    """Dataclass of render resources with color."""

    color: RGBAColor


@dataclass
class Polygon(ColorRenderResource):
    """Dataclass containing data for drawing polygon."""

    points: Iterable[Vector]


@dataclass
class Line(ColorRenderResource):
    """Dataclass containing data for drawing line."""

    start_point: Vector
    end_point: Vector


@dataclass
class Circle(ColorRenderResource):
    """Dataclass containing data for drawing circle."""

    radius: int | float


@dataclass
class Rectangle(ColorRenderResource):
    """Dataclass containing data for drawing rectangle."""

    width: int | float
    height: int | float



from abc import ABC, abstractmethod
from dataclasses import dataclass
from math import sqrt, fabs, degrees, acos, cos, asin, sin, radians
from functools import lru_cache, wraps, cached_property, reduce
from typing import Iterable, Callable, Union, Generator, Self

from beautiful_repr import StylizedMixin, Field, TemplateFormatter, parse_length
from pyoverload import overload

from sim32.interfaces import IUpdatable, IZone, IZoneFactory
from sim32.errors.geometry_errors import *
from sim32.tools import *


class DegreeMeasure:
    """Class for holding degrees in normalized form and manipulations with them."""

    __slots__ = ('__degrees')

    def __init__(self, degrees: int | float):
        self.__degrees = self._bring_number_into_degrees(degrees)

    @property
    def degrees(self) -> int | float:
        """Property for unprotected degree form."""

        return self.__degrees

    def __int__(self) -> int:
        return int(self.degrees)

    def __float__(self) -> float:
        return float(self.degrees)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.degrees})"

    def __hash__(self) -> int:
        return self.degrees

    def _degree_measure_creation_from_degrees(
        method: Callable[[Self, any], int | float]
    ) -> Callable[[any], Self]:
        """Method decorator for native creation of DegreemMeasure from output numeric degrees."""

        @wraps(method)
        def wrapper(self: Self, *args, **kwargs) -> Self:
            return self.__class__(method(self, *args, **kwargs))

        return wrapper

    def _interpret_input_measure_in_degrees(
        method: Callable[[int | float], any]
    ) -> Callable[[Union[int, float, Self]], any]:
        """Decorator for methods that converts the input DegreeMeasure to its bare degrees."""

        @wraps(method)
        def wrapper(
            self,
            other: Union[int, float, Self],
            *args,
            **kwargs
        ) -> any:
            return method(
                self,
                self._get_number_from_degrees_or_number(other),
                *args,
                **kwargs
            )

        return wrapper

    @_interpret_input_measure_in_degrees
    def __eq__(self, number: int | float) -> bool:
        return self.degrees == number

    @_interpret_input_measure_in_degrees
    def __ne__(self, number: int | float) -> bool:
        return self.degrees != number

    @_interpret_input_measure_in_degrees
    def __lt__(self, number: int | float) -> bool:
        return self.degrees < number

    def __le__(self, number: int | float | Self) -> bool:
        return self < number or self == number

    @_interpret_input_measure_in_degrees
    def __gt__(self, number: int | float) -> bool:
        return self.degrees > number

    def __ge__(self, number: int | float | Self) -> bool:
        return self > number or self == number

    @_degree_measure_creation_from_degrees
    @_interpret_input_measure_in_degrees
    def __add__(self, number: int | float) -> int | float:
        return self.degrees + number

    def __radd__(self, number: int | float) -> Self:
        return self + number

    @_degree_measure_creation_from_degrees
    @_interpret_input_measure_in_degrees
    def __sub__(self, number: int | float) -> int | float:
        return self.degrees - number

    def __rsub__(self, number: int | float) -> Self:
        return -(self) + number

    @_degree_measure_creation_from_degrees
    @_interpret_input_measure_in_degrees
    def __mul__(self, number: int | float) -> int | float:
        return self.degrees * number

    def __rmul__(self, number: int | float) -> Self:
        return self * number

    @_degree_measure_creation_from_degrees
    @_interpret_input_measure_in_degrees
    def __truediv__(self, number: int | float) -> int | float:
        return self.degrees / number

    @_degree_measure_creation_from_degrees
    @_interpret_input_measure_in_degrees
    def __floordiv__(self, number: int | float) -> int | float:
        return (self.degrees / number) // 1

    @_degree_measure_creation_from_degrees
    def __neg__(self) -> Self:
        return self.degrees * -1

    def _get_number_from_degrees_or_number(
        cls,
        number_or_degrees: Union[int, float, Self]
    ) -> int | float:
        """
        Method for natively converting the input DegreeMeasure to its bare
        numeric form, or non-converting if a number is entered.
        """

        return (
            number_or_degrees.degrees if isinstance(number_or_degrees, DegreeMeasure)
            else number_or_degrees
        )

    @staticmethod
    def _bring_number_into_degrees(number: int | float) -> int | float:
        """Method for resetting a number on overflow greater than 359."""

        return number - (number // 360)*360


@dataclass(repr=False, frozen=True)
class AxisPlaneDegrees:
    """Dataclass for positioning degrees on an axial plane."""

    first_axis: int
    second_axis: int
    degrees: DegreeMeasure

    def __post_init__(self) -> None:
        if self.first_axis == self.second_axis:
            raise AxisPlaneDegreesError(f"{self.__class__.__name__} must be on two axes, not one ({self.first_axis})")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self.axes)[1:-1]}, degrees={self.degrees.degrees})"

    @cached_property
    def axes(self) -> tuple[int, int]:
        """Property representing axes in collection format."""

        return (self.first_axis, self.second_axis)

    def is_on_same_plane_with(self, axis_degrees: Self) -> bool:
        """Method for comparing axes of two AxisPlaneDegrees."""

        return frozenset(self.axes) == frozenset(axis_degrees.axes)

    def get_external(self) -> Self:
        """Method for getting the other degrees located in the same plane."""

        return self.__class__(self.first_axis, self.second_axis, -self.degrees)


@dataclass(repr=False, frozen=True)
class DegreeArea(AxisPlaneDegrees):
    """Class representing degrees on a plane in a certain zone."""

    shift_degrees: DegreeMeasure

    def __repr__(self) -> str:
        return "{class_name}({axes}, degrees={degree_diapason})".format(
            class_name=self.__class__.__name__,
            axes=str(self.axes)[1:-1],
            degree_diapason=f"({self.shift_degrees.degrees} ~ {self.border_degrees.degrees})"
        )

    def __contains__(self, degrees: int | float | DegreeMeasure) -> bool:
        return self.is_degrees_inside(degrees)

    @cached_property
    def border_degrees(self) -> DegreeMeasure:
        """Property for defining the top border of the area."""

        return self.degrees + self.shift_degrees

    @cached_property
    def is_empty(self) -> bool:
        """Property for determining the emptiness of an area."""

        return self.shift_degrees == self.border_degrees

    def is_degrees_inside(self, degrees: int | float | DegreeMeasure) -> bool:
        """Method for determining the presence of degrees in an area."""

        if self.is_empty:
            return False

        is_degrees_in_diapason = degrees in self._diapason

        return (
            is_degrees_in_diapason
            if self.shift_degrees <= self.border_degrees
            else (
                not is_degrees_in_diapason
                or degrees == self.shift_degrees
                or degrees == self.border_degrees
            )
        )

    def get_external(self) -> Self:
        return self.__class__(
            self.first_axis,
            self.second_axis,
            -self.degrees,
            self.border_degrees
        )

    @cached_property
    def _diapason(self) -> Diapason:
        """Degree range rough shape property."""

        return Diapason(
            self.shift_degrees,
            self.border_degrees,
            is_end_inclusive=True
        )


class Vector:
    """Class for manipulating vectors. Are not strict to the number of measurements."""

    def __init__(self, coordinates: Iterable[float | int] = tuple()):
        self.__coordinates = tuple(coordinates)

    @property
    def coordinates(self) -> tuple[int | float]:
        """Vector Coordinate Property."""

        return self.__coordinates

    @cached_property
    def length(self) -> float:
        """Vector length property."""

        return sqrt(sum(coordinate**2 for coordinate in self.coordinates))

    @cached_property
    def degrees(self) -> tuple[AxisPlaneDegrees]:
        """Property of degrees of a vector sliced on a plane."""

        perpendicular_vector = Vector((1, ))

        return tuple(
            AxisPlaneDegrees(
                first_axis,
                second_axis,
                self.__class__((
                    self.coordinates[first_axis],
                    self.coordinates[second_axis]
                )).get_degrees_between(
                    perpendicular_vector,
                    0 > self.coordinates[second_axis]
                )
            )
            for first_axis in range(len(self.coordinates))
            for second_axis in range(first_axis + 1, len(self.coordinates))
            if any(map(
                lambda coordinate: coordinate != 0,
                (self.coordinates[first_axis], self.coordinates[second_axis])
            ))
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(tuple(self.coordinates))[1:-1]})"

    def __hash__(self) -> int:
        return hash(self.coordinates)

    def __eq__(self, other: Self) -> Self:
        return self.coordinates == other.coordinates

    @lru_cache(maxsize=8192)
    def __add__(self, other: Self) -> Self:
        return self.__class__(
            tuple(map(
                lambda first, second: first + second,
                *(
                    vector.coordinates
                    for vector in self.get_mutually_normalized((self, other))
                )
            ))
        )

    def __sub__(self, other: Self) -> Self:
        return self + (-other)

    @lru_cache(maxsize=4096)
    def __mul__(self, other: Union[int, float, Self]) -> Self:
        return (
            self.get_scalar_by(other) if isinstance(other, Vector)
            else self.get_multiplied_by_number(other)
        )

    def __rmul__(self, number: int | float) -> Self:
        return self * number

    def __truediv__(self, number: int | float) -> Self:
        return self*(1 / number)

    def __floordiv__(self, number: int | float) -> Self:
        return self.__class__(
            tuple(int(coordinate) for coordinate in (self / number).coordinates)
        )

    def __neg__(self) -> Self:
        return self.get_reflected_by_axes()

    def __len__(self) -> int:
        return len(self.coordinates)

    @lru_cache(maxsize=1024)
    def get_normalized_to_measurements(
        self,
        number_of_measurements: int,
        default_measurement_point: int | float = 0
    ) -> Self:
        """
        Method for getting a vector expanded or reduced to a certain number of
        dimensions.
        """

        measurement_difference = number_of_measurements - len(self.coordinates)

        return self.__class__(
            self.coordinates + (default_measurement_point,)*measurement_difference if measurement_difference > 0
            else self.coordinates[:number_of_measurements if number_of_measurements >= 0 else 0]
        )

    @lru_cache(maxsize=128)
    def get_reflected_by_axes(
        self,
        axis_indexes: Iterable[int] | None = None
    ) -> Self:
        """Method for getting a vector unfolded in a plane."""

        if axis_indexes is None:
            axis_indexes = range(len(self.coordinates))

        return self.__class__(tuple(
            coordinate * (-1 if coordinate_index in axis_indexes else 1)
            for coordinate_index, coordinate in enumerate(self.coordinates)
        ))

    @lru_cache(maxsize=128)
    def get_reduced_to_length(self, length: int | float) -> Self:
        """Method for getting a similar vector but with an input length."""

        if self.length == 0:
            raise VectorError("Vector with length == 0 can't be lengthened")

        return (self / self.length) * length

    def get_rotated_many_times_by(self, axis_degree_measures: Iterable[AxisPlaneDegrees]) -> Self:
        """Method for getting a rotated vector along many axes many times."""

        result_vector = self

        for axis_degree_measure in axis_degree_measures:
            result_vector = result_vector.get_rotated_by(axis_degree_measure)

        return result_vector

    def get_rotated_by(self, axes_degrees: AxisPlaneDegrees) -> Self:
        """Method for getting a vector rotated on a plane."""

        number_of_measurements = max(axes_degrees.axes) + 1
        reduced_vector = (
            self.get_normalized_to_measurements(number_of_measurements)
            if len(self.coordinates) < number_of_measurements else self
        )

        axes_section_vector = self.__class__((
            reduced_vector.coordinates[axes_degrees.first_axis],
            reduced_vector.coordinates[axes_degrees.second_axis]
        ))

        if all(coordinate == 0 for coordinate in axes_section_vector.coordinates):
            return self

        coordinates = list(reduced_vector.coordinates)

        reduced_axes_section_vector = axes_section_vector.get_reduced_to_length(1)

        coordinates[axes_degrees.first_axis] = axes_section_vector.length * cos(radians(
                axes_section_vector.degrees[0].degrees
                + axes_degrees.degrees
        ))

        coordinates[axes_degrees.second_axis] = axes_section_vector.length * sin(radians(
                axes_section_vector.degrees[0].degrees
                + axes_degrees.degrees
        ))

        return self.__class__(coordinates)

    def get_rounded_by(self, rounder: NumberRounder) -> Self:
        """Method for getting rounded vector with rounding input implementation."""

        return self.__class__(tuple(
            rounder(coordinate)
            for coordinate in self.coordinates
        ))

    def get_multiplied_by_number(self, number: int | float) -> Self:
        """Method for getting a vector multiplied by a number."""

        return self.__class__(
            tuple(number * coordinate for coordinate in self.coordinates)
        )

    def get_scalar_by(self, vector: Self) -> int | float:
        """Method to get a scalar between two vectors."""

        return sum(tuple(map(
            lambda first, second: first * second,
            *(
                normalized_vector.coordinates
                for normalized_vector in self.get_mutually_normalized((self, vector))
            )
        )))

    def get_degrees_between(self, vector: Self, is_external: bool = False) -> DegreeMeasure:
        """Method for getting angle degrees between two vectors."""

        return DegreeMeasure(degrees(acos(
            (self * vector) / (self.length * vector.length)
        ))) * (-1 if is_external else 1)

    @staticmethod
    def get_mutually_normalized(vectors: Iterable[Self]) -> tuple[Self]:
        """Method for getting vectors in the same dimensions."""

        maximum_number_of_measurements = max((len(vector.coordinates) for vector in vectors))

        return tuple(
            vector.get_normalized_to_measurements(maximum_number_of_measurements)
            for vector in vectors
        )

    @classmethod
    def create_by_degrees(cls, length: int | float, axis_degree_measures: Iterable[AxisPlaneDegrees]) -> Self:
        """
        Method for creating a vector of an input length unfolded along the
        degrees of input certain planes.

        Doesn't guarantee the creation of a vector with degrees contradicting
        the input.
        """

        fill_axis = axis_degree_measures[0].first_axis if len(axis_degree_measures) else 0

        return cls(
            (0, )*fill_axis + (length, )
        ).get_rotated_many_times_by(axis_degree_measures)


@dataclass(repr=False)
class PositionVector:
    """Dataclass to emulate a vector with a specific start and end."""

    start_point: Vector
    end_point: Vector

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(from {self.start_point} to {self.end_point})"

    @property
    def virtual_vector(self) -> Vector:
        """Property for immediate vector lying between start and end."""

        return self.end_point - self.start_point

    def get_rounded_by(self, rounder: NumberRounder) -> Self:
        """Method for rounding vectors annotating start and end."""

        return self.__class__(
            self.start_point.get_rounded_by(rounder),
            self.end_point.get_rounded_by(rounder)
        )


class IPointChanger(ABC):
    """
    Vector changer interface for changing a vector without dependencies on a
    specific change implementation.
    """

    @abstractmethod
    def __call__(self, point: Vector) -> Vector:
        """Method for getting a vector modified according to the similarity of the input."""


class DynamicTransporter(IPointChanger):
    """
    PointChanger class which returns a vector sum of the input and that contained in
    the changer itself.

    Abstracts addition.
    """

    def __init__(self, shift: Vector):
        self.shift = shift

    def __call__(self, point: Vector) -> Vector:
        return point + self.shift


class PointRotator(StylizedMixin, IPointChanger):
    """
    PointChanger class returning the rotated analog of the input vector.

    Abstracts rotation.
    Rotates the vector from the center_point vector by axis_degree_measures degrees.
    """

    _repr_fields = (
        Field('center_point', formatter=TemplateFormatter('about {value}')),
        Field(
            'axis_degree_measures',
            formatter=TemplateFormatter('{value}'),
            value_transformer=lambda value: str(list(value))[1:-1] if len(value) > 0 else value
        )
    )

    def __init__(self, axis_degree_measures: Iterable[AxisPlaneDegrees], center_point: Vector = Vector()):
        self.axis_degree_measures = tuple(axis_degree_measures)
        self.center_point = center_point

    def __call__(self, point: Vector) -> Vector:
        return reduce(
            lambda result_point, axis_degree_measure: (
                (result_point - self.center_point).get_rotated_by(axis_degree_measure)
                + self.center_point
            ),
            (point, *self.axis_degree_measures)
        )


class VectorDivider(Divider, StylizedMixin):
    """Class dividing a vector into points."""

    _repr_fields = (Field('distance_between_points'), )

    def __init__(self, distance_between_points: int | float, rounder: NumberRounder):
        self.distance_between_points = distance_between_points
        self.rounder = rounder

    def is_possible_to_divide(self, data: Vector) -> Report:
        return Report.create_error_report(
            UnableToDivideVectorIntoPointsError(
                f"Can't divide vector {data} into points with length 0"
            )
        ) if data.virtual_vector.length == 0 else super().is_possible_to_divide(data)

    def _divide(self, vector: PositionVector) -> frozenset[Vector]:
        distance_factor = self.distance_between_points / vector.virtual_vector.length

        vector_to_next_point = Vector(tuple(
            coordinate * distance_factor for coordinate in vector.virtual_vector.coordinates
        ))

        return self.__create_points(
            vector.start_point,
            vector.virtual_vector.length / vector_to_next_point.length,
            vector_to_next_point
        )

    def __create_points(
        self,
        start_point: Vector,
        number_of_points_to_create: int,
        vector_to_next_point: Vector
    ) -> frozenset[Vector]:
        created_points = [start_point]

        for created_point_index in range(1, int(number_of_points_to_create) + 1):
            created_points.append(
                created_points[created_point_index - 1] + vector_to_next_point
            )

        return frozenset(
            point.get_rounded_by(self.rounder) for point in created_points
        )


class Figure(IZone, ABC):
    """
    Base zone class.

    Template-wise implements finding vectors in the zone by dividing them into
    points and working with them already.
    """

    _vector_divider_factory: Callable[['Line'], VectorDivider] = (
        lambda _: VectorDivider(0.1, ShiftNumberRounder(AccurateNumberRounder(), 1))
    )

    def __init__(self):
        self._vector_divider = self._vector_divider_factory()

    @overload
    def __contains__(self, point: Vector) -> bool:
        return self.is_point_inside(point)

    @overload
    def __contains__(self, vector: PositionVector) -> bool:
        return self.is_vector_passes(vector)

    def is_vector_passes(self, vector: PositionVector) -> bool:
        return any(
            self.is_point_inside(point)
            for point in self._vector_divider(vector)
        )

    def is_vector_entered(self, vector: PositionVector) -> bool:
        return self.is_point_inside(
            vector.end_point
        )


class AxisZone(Figure, StylizedMixin):
    """
    Class that implements a zone representation as a space between two points.

    Similar to a rectangle, but only because of its implementation of finding a
    point in it compared to its coordinates with its own diapasons of valid
    coordinates of certain axes.
    """

    _repr_fields = Field('axis_diapasons', formatter=lambda value, _: f"{str(value)[1:-1]}"),

    def __init__(self, first_point: Vector, second_point: Vector):
        super().__init__()
        self.__first_point = first_point
        self.__second_point = second_point

        self._update()

    @property
    def first_point(self) -> Vector:
        return self.__first_point

    @first_point.setter
    def first_point(self, new_point: Vector) -> Vector:
        self.__first_point = new_point
        self._update()

    @property
    def second_point(self) -> Vector:
        return self.__second_point

    @second_point.setter
    def second_point(self, new_point: Vector) -> Vector:
        self.__second_point = new_point
        self._update()

    @property
    def size(self) -> tuple[int | float]:
        """Virtual rectangle size property."""

        return self.__size

    @property
    def axis_diapasons(self) -> tuple[Diapason]:
        """Property of diapasons of available coordinates along the axes."""

        return self.__axis_diapasons

    def move_by(self, point_changer: IPointChanger) -> None:
        self.first_point = point_changer(self.first_point)
        self.second_point = point_changer(self.second_point)

    def is_point_inside(self, point: Vector) -> bool:
        return all(
            coordinate in self.axis_diapasons[axis]
            for axis, coordinate in enumerate(
                point.get_normalized_to_measurements(len(self.size)).coordinates
            )
        )

    @classmethod
    def create_with_generated_points_by(
        cls,
        center_point: Vector,
        size: Iterable[int | float]
    ) -> Self:
        """Creation method using point generation by size and center point."""

        vector_to_extreme_point = Vector(size) / 2

        return cls(
            center_point + vector_to_extreme_point,
            center_point - vector_to_extreme_point
        )

    @classmethod
    def create_as_square(
        cls,
        center_point: Vector,
        side_length: int | float,
        number_of_measurements: int
    ) -> Self:
        """
        Method of creating by the shape of a square using the generation of
        points by the center point and the length of the edge of the square.
        """

        return cls.create_with_generated_points_by(
            center_point,
            (side_length, ) * number_of_measurements
        )

    def _update(self) -> None:
        """Method for updating the whole body of a zone."""

        self.__first_point, self.__second_point = Vector.get_mutually_normalized((
            self.__first_point,
            self.__second_point
        ))

        self.__axis_diapasons = tuple(
            Diapason(first_point_coordinate, second_point_coordinate, True)
            for first_point_coordinate, second_point_coordinate  in zip(
                self.__first_point.coordinates,
                self.__second_point.coordinates
            )
        )

        self.__size = tuple(
            axis_diapason.end - axis_diapason.start
            for axis_diapason in self.__axis_diapasons
        )


class Angle(Figure, StylizedMixin):
    """Class of providing a zone of an angle originating from some point."""

    _repr_fields = Field('center_point'),

    def __init__(self, center_point: PositionVector, degree_areas: Iterable[DegreeArea]):
        self._center_point = center_point
        self._degree_areas = tuple(degree_areas)

        self.update_by_points(self.create_ray_vertices_by(1))

    @property
    def center_point(self) -> Vector:
        """Center point property."""

        return self._center_point

    @property
    def degree_areas(self) -> tuple[DegreeArea]:
        """Angle degree property."""

        return self._degree_areas

    def move_by(self, point_changer: IPointChanger) -> None:
        self._center_point = point_changer(self._center_point)

        self.update_by_points(tuple(
            point_changer(vertex) for vertex in self.create_ray_vertices_by(1)
        ))

    def create_ray_vertices_by(self, length: int | float) -> frozenset[Vector]:
        """Method for creating vectors with a fixed length from center point."""

        return frozenset(
            self.center_point + Vector.create_by_degrees(
                length,
                (AxisPlaneDegrees(
                        degree_area.first_axis,
                        degree_area.second_axis,
                        degrees
                ), )
            )
            for degree_area in self._degree_areas
            for degrees in (degree_area.shift_degrees, degree_area.border_degrees)
        )


    def update_by_points(self, points: Iterable[Vector]) -> None:
        self._degree_areas = tuple(self.__create_degree_areas_from(
            tuple(get_collection_with_reduced_nesting_level_by(
                1,
                (
                    (point - self._center_point).degrees
                    for point in Vector.get_mutually_normalized(points)
                )
            ))
        ))

    def become_external(self) -> None:
        """Angle reversal method."""

        self._degree_areas = tuple(
            degree_area.get_external()
            for degree_area in self._degree_areas
        )

    def is_point_inside(self, point: Vector) -> bool:
        return point == self._center_point or not any(
            degree_measure.degrees.degrees not in self.get_degree_area_by_axes(
                degree_measure.first_axis,
                degree_measure.second_axis
            )
            for degree_measure in (point - self._center_point).degrees
        )

    def get_degree_area_by_axes(self, first_axis: int, second_axis: int) -> DegreeArea:
        """Vector to get the axial degrees of the angle along the input plane."""

        for degree_area in self.degree_areas:
            if degree_area.first_axis == first_axis and degree_area.second_axis == second_axis:
                return degree_area
        else:
            return DegreeArea(first_axis, second_axis, DegreeMeasure(0), DegreeMeasure(0))

    @classmethod
    def created_by_points(cls, center_point: PositionVector, points: Iterable[Vector]) -> Self:
        """
        Method for creating from a set of points. Creates an angle capable of
        accommodating all input.
        """

        angle = cls(center_point, tuple())
        angle.update_by_points(points)

        return angle

    def __create_degree_areas_from(self, axis_degree_measures: Iterable[AxisPlaneDegrees]) -> Generator[DegreeArea, any, None]:
        """
        Method for creating degree areas from multiple axial degrees.
        Is a generator.
        """

        max_axes = 1 + max(get_collection_with_reduced_nesting_level_by(
            1,
            (point_degree_measure.axes for point_degree_measure in axis_degree_measures)
        ))

        for first_axis in range(max_axes):
            for second_axis in range(first_axis + 1, max_axes):
                degree_multitude = frozenset(
                    degree_measure.degrees.degrees
                    for degree_measure in axis_degree_measures
                    if (
                        degree_measure.first_axis == first_axis
                        and degree_measure.second_axis == second_axis
                    )
                )

                min_degrees, max_degrees = min(degree_multitude), max(degree_multitude)

                yield DegreeArea(
                    first_axis,
                    second_axis,
                    DegreeMeasure(max_degrees - min_degrees),
                    DegreeMeasure(min_degrees)
                )


class Site(Figure):
    """Class representation of a point with a zone interface."""

    def __init__(self, point: Vector):
        self.point = point

    def move_by(self, point_changer: IPointChanger) -> None:
        self.point = point_changer(self.point)

    def is_point_inside(self, point: Vector) -> bool:
        return self.point == point


class CompositeFigure(Figure, StylizedMixin):
    """Zone proxy class consisting of multiple zones."""

    _repr_fields = (
        Field(
            'main_figures',
            value_getter=parse_length,
            formatter=TemplateFormatter("{value} main figures")
        ),
        Field(
            'subtraction_figures',
            value_getter=parse_length,
            formatter=TemplateFormatter("{value} subtraction figures")
        )
    )

    def __init__(
        self,
        main_figures: Iterable[Figure],
        subtraction_figures: Iterable[Figure] = tuple()
    ):
        super().__init__()
        self.main_figures = set(main_figures)
        self.subtraction_figures = set(subtraction_figures)

    def move_by(self, point_changer: IPointChanger) -> None:
        for figure in (*self.main_figures, *self.subtraction_figures):
            figure.move_by(point_changer)

    def is_point_inside(self, point: Vector) -> bool:
        return (
            any(figure.is_point_inside(point) for figure in self.main_figures)
            if all(not figure.is_point_inside(point) for figure in self.subtraction_figures)
            else False
        )


class Line(Figure, StylizedMixin):
    """Position vector representation class with zone interface."""

    _repr_fields = (
        Field(
            value_getter=lambda line, _: (line.first_point, line.second_point),
            formatter=lambda values, _: f"between {values[0]} and {values[1]}"
        ),
    )

    def __init__(self, first_point: Vector, second_point: Vector):
        super().__init__()

        self._rounder = self._vector_divider.rounder

        self.__first_point = first_point
        self.__second_point = second_point

        self._update_points()

    @property
    def first_point(self) -> Vector:
        return self.__first_point

    @first_point.setter
    def first_point(self, new_point: Vector) -> None:
        self.__first_point = new_point
        self._update_points()

    @property
    def second_point(self) -> Vector:
        return self.__second_point

    @second_point.setter
    def second_point(self, new_point: Vector) -> None:
        self.__second_point = new_point
        self._update_points()

    @property
    def all_available_points(self) -> tuple[Vector]:
        """Class of all possible line points."""

        return self.__all_available_points

    def move_by(self, point_changer: IPointChanger) -> None:
        self.__first_point, self.__second_point = map(
            point_changer, (self.first_point, self.second_point)
        )

        self._update_points()

    def is_vector_passes(self, vector: PositionVector) -> bool:
        return super().is_vector_passes(vector) if (
            vector.start_point in self.__proposed_location_area or
            vector.end_point in self.__proposed_location_area or
            vector.end_point - vector.virtual_vector*0.5 in self.__proposed_location_area
        ) else False

    def is_point_inside(self, point: Vector) -> bool:
        return (
            point.get_rounded_by(self._rounder) in self.__all_available_points
            if not point in self.__proposed_location_area else True
        )

    def _update_points(self) -> None:
        """Method for updating all possible points."""

        self.__first_point, self.__second_point = map(
            lambda vector: vector.get_rounded_by(self._rounder),
            (self.__first_point, self.__second_point)
        )

        self.__all_available_points = self._vector_divider(
            PositionVector(self.first_point, self.second_point)
        )
        self.__proposed_location_area = AxisZone(self.first_point, self.second_point)


class Polygon(Figure, StrictToStateMixin, StylizedMixin):
    """Polygon Face (!) Zone Class."""

    _repr_fields = (
        Field(
            'summits',
            value_getter=parse_length,
            formatter=TemplateFormatter("{value} summits")
        ),
    )
    _line_factory: Callable[[Vector, Vector], Line] = Line
    _report_analyzer = ReportAnalyzer(
        (BadReportHandler(FigureIsNotCorrect, "Polygon not viable"), )
    )

    def __init__(self, points: Iterable[Vector]):
        super().__init__()
        self._update_lines_by(points)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self.summits)} summit{'s' if len(self.summits) > 0 else ''})"

    @property
    def summits(self) -> tuple[Vector]:
        return self.__summits

    def move_by(self, point_changer: IPointChanger) -> None:
        self._update_lines_by(
            tuple(map(point_changer, self.summits))
        )
        self._check_state_errors()

    def is_point_inside(self, point: Vector) -> bool:
        return any(line.is_point_inside(point) for line in self._lines)

    def _is_correct(self) -> Report:
        number_of_measurements = max(
            map(lambda point: len(point.coordinates), self.summits)
        )

        if len(self.summits) <= number_of_measurements:
            return Report.create_error_report(FigureIsNotClosedError(
                f"{number_of_measurements}D figure must contain more than {number_of_measurements} links for closure"
            ))
        else:
            return Report(True)

    def _update_lines_by(self, points: Iterable[Vector]) -> tuple[Line]:
        self._lines = tuple(
            self._line_factory(
                points[point_index - 1],
                points[point_index]
            )
            for point_index in range(len(points))
        )

        self.__summits = tuple(line.first_point for line in self._lines)
        self._check_state_errors()


class Circle(Figure, StylizedMixin):
    """Zone class of a circle or its multidimensional variations."""

    _repr_fields = (Field('radius'), Field('center_point'))

    def __init__(self, center_point: Vector, radius: int | float):
        super().__init__()
        self.center_point = center_point
        self.radius = radius

    def move_by(self, point_changer: IPointChanger) -> None:
        self.center_point = point_changer(self.center_point)

    def is_point_inside(self, point: Vector) -> bool:
        return (self.center_point - point).length <= self.radius



from typing import NewType
from sys import version_info
from os import *


is_supports_self_annotation = version_info >= (3, 11)

if not is_supports_self_annotation:
    Self: NewType = object

current_path = path.dirname(path.abspath(__file__))
current_directory = path.basename(current_path)

file_names_of_module = (
    file_name for file_name in listdir(current_path)
    if '.py' == file_name[-3:]
)

ignore_fiele_names = ('__init__.py', )

for file_name in (
    file_name
    for file_name in file_names_of_module
    if file_name not in ignore_fiele_names
):
    exec(f"from {current_directory}.{file_name.split('.')[0]} import *")

if not is_supports_self_annotation:
    del Self



from abc import ABC, abstractmethod
from typing import Iterable, Callable, Optional, Self, NamedTuple

from beautiful_repr import StylizedMixin, Field

from sim32.interfaces import *
from sim32.renders import ResourcePack, RenderActivator, IRender
from sim32.errors.core_errors import *
from sim32.tools import *
from sim32.geometry import Vector, Figure, Site, DynamicTransporter, IPointChanger


class IProcessState(IUpdatable, ABC):
    """Interface for public process behavior."""

    @property
    def process(self) -> 'Process':
        """Property for process that has this state."""

    @abstractmethod
    def get_next_state(self) -> Optional[Self]:
        """Property of the next state of the process."""

    @abstractmethod
    def is_compelling_to_handle(self) -> bool:
        """Flag property defining the internal behavior of the process."""

    @abstractmethod
    def is_valid(self) -> Report:
        """
        Property denoting the validity of the state of the process for further
        exploitation.
        """


class ProcessState(StrictToStateMixin, IProcessState, ABC):
    """
    Basic implementation of the ProcessState interface.

    Raises an error when attempting to call with an invalid state.
    """

    _state_report_analyzer = ReportAnalyzer((BadReportHandler(
        ProcessStateIsNotValidError,
        "Process state is not valid to update"
    ), ))

    def __init__(self, process: 'Process'):
        self.__process = process

    def __hash__(self) -> int:
        return id(self.__class__)

    @property
    def process(self) -> 'Process':
        return self.__process

    def update(self) -> None:
        self._check_state_errors()
        self._handle()

    @abstractmethod
    def _handle(self) -> None:
        """Method for handling the public state of the process."""

    def _is_correct(self) -> Report:
        return self.is_valid()


class CompletedProcessState(ProcessState):
    """Completed process state class. Raises an error during updating."""

    is_compelling_to_handle = False

    def get_next_state(self) -> None:
        return None

    def is_valid(self) -> Report:
        return Report(True)

    def _handle(self) -> None:
        raise ProcessAlreadyCompletedError(
            f"Process {self.process} has completed its life cycle"
        )


class ActiveProcessState(ProcessState):
    """
    Standard process state class that allows an internal state to flow without
    the need for an public one.
    """

    is_compelling_to_handle = True

    def get_next_state(self) -> None:
        return None

    def is_valid(self) -> Report:
        return Report(True)

    def _handle(self) -> None:
        pass


class NewStateByValidationProcessStateMixin(IProcessState, ABC):
    """
    Process state mixin that defines a new state when the current one is invalid.

    Provides for the creation of the next process using the _new_state_factory
    attribute.
    """

    _new_state_factory: Callable[['Process'], ProcessState | None] = CustomFactory(ActiveProcessState)

    def get_next_state(self) -> ProcessState | None:
        return self._new_state_factory(self.process) if not self.is_valid() else None


class SleepProcessState(ProcessState, NewStateByValidationProcessStateMixin):
    """
    Process state class the freezing action of the internal state of the process
    for the specified number of runs of the update method.
    """

    is_compelling_to_handle = False

    def __init__(
        self,
        process: 'Process',
        ticks_to_activate: int | float,
        tick_factor: int | float = 1
    ):
        super().__init__(process)
        self.ticks_to_activate = ticks_to_activate
        self.tick = 1 * tick_factor

    def __hash__(self) -> int:
        return id(self)

    def is_valid(self) -> Report:
        return Report.create_error_report(
            ProcessIsNoLongerSleepingError(f"Process {self.process} no longer sleeps")
        ) if self.ticks_to_activate <= 0 else Report(True)

    def _handle(self) -> None:
        self.ticks_to_activate -= self.tick


class FlagProcessState(ProcessState, NewStateByValidationProcessStateMixin):
    """
    Process state class that doesn't handle anything but annotates handling to
    something else.
    """

    is_compelling_to_handle = True
    _is_standing: bool = False

    def is_valid(self) -> Report:
        return Report(self._is_standing)

    @classmethod
    def create_flag_state(
        cls,
        name: str,
        is_standing: bool = False,
        bases: Iterable[type] = tuple(),
        attributes: dict = dict()
    ) -> Self:
        """Process state flag class dynamic creation method."""

        return type(
            name,
            bases + (cls, ),
            {'is_standing': is_standing} | attributes
        )

    def _handle(self) -> None:
        pass


class IProcess(IUpdatable, ABC):
    """
    Process class, which is the removal of the processing logic of something
    into an object to systematize work with this logic.

    It has two states: internal - which is the logic rendered into the object
    and public - the state imposed from outside to control this process. Public
    state is expressed by an object and lies in the state attribute.
    """

    state: IProcessState | None

    @property
    @abstractmethod
    def original_process(self) -> Self:
        """
        Crutch property that ensures the availability of the original process
        behind a layer of proxy processes.
        """

    @property
    @abstractmethod
    def participants(self) -> tuple:
        """Property of objects involved in the process."""

    @abstractmethod
    def start(self) -> None:
        """Process start method."""


class Process(StrictToStateMixin, IProcess, ABC):
    _state_report_analyzer = ReportAnalyzer((BadReportHandler(
        ProcessError,
        "Process is not valid"
    ), ))

    state = None

    def __init__(self):
        self._check_state_errors()

    @property
    def original_process(self) -> IProcess:
        return self

    def start(self) -> None:
        self.state = ActiveProcessState(self)

    def update(self) -> None:
        if not self.state:
            self.start()

        while True:
            if self.state.is_valid():
                self.state.update()

            old_state = self.state
            self.__reset_state()

            if hash(old_state) == hash(self.state):
                break

        if self.state.is_compelling_to_handle:
            self._handle()

    @abstractmethod
    def _handle(self) -> None:
        """Process internal state method."""

    def _get_next_state(self) -> ProcessState | None:
        return None

    def _is_correct(self) -> Report:
        return Report(True)

    def __reset_state(self) -> None:
        """Method for updating its public state."""

        next_state = self.state.get_next_state()

        if next_state is None:
            next_state = self._get_next_state()

        if next_state:
            self.state = next_state


class ProxyProcess(IProcess, ABC):
    """Process class that changes the logic of another process."""

    def __init__(self, process: IProcess):
        self._process = process

    @property
    def process(self) -> IProcess:
        return self._process

    @property
    def original_process(self) -> IProcess:
        current_process = self._process

        while isinstance(current_process, ProxyProcess):
            current_process = current_process.process

        return current_process

    @property
    def state(self) -> IProcessState | None:
        return self.process.state

    @state.setter
    def state(self, new_state: IProcessState | None) -> None:
        self.process.state = new_state

    @property
    def participants(self) -> tuple:
        return self._process.participants

    def start(self) -> None:
        self.process.start()

    def update(self) -> None:
        self._process.update()


class StrictToParticipantsProcess(Process, ABC):
    """Process class that has strict restrictions on the states of its participants."""

    def _is_correct(self) -> Report:
        self.is_support_participants(self.participants)

    @classmethod
    @abstractmethod
    def is_support_participants(cls, participants: Iterable) -> Report:
        """
        Method for validating the state of participants without taking into
        account the state of the process itself.
        """


class ManyPassProcess(Process, ABC):
    """
    Process class that natively ends after a certain number of updates.

    Сertain number of updates is specified by the _passes attribute.
    """

    _passes: int

    def update(self) -> None:
        self._passes -= 1
        super().update()

    def _get_next_state(self) -> ProcessState | None:
        return CompletedProcessState(self) if self._passes <= 0 else None


class WorldProcess(Process, ABC):
    """
    Process world handling class.
    Before starting, the world attribute must be filled with the defined world.
    """

    world: Optional['World'] = None

    def start(self) -> None:
        if not self.world:
            raise WorldProcessError(f"World process {self} has no world")

        super().start()


class Event(Process, ABC):
    """Process class with homogeneous unrestricted participants."""

    def __init__(self, participants: Iterable[IUpdatable]):
        self.__participants = tuple(participants)
        super().__init__()

    @property
    def participants(self) -> tuple:
        return self.__participants


class FocusedEvent(Event, ABC):
    """Event class that handles each of its participants in the same way."""

    def _handle(self) -> None:
        for participant in self.participants:
            self._handle_participant(participant)

    @abstractmethod
    def _handle_participant(self, participant: IUpdatable) -> None:
        """Participant handling method applied to each participant."""


class UnitSpawnProcess(FocusedEvent, WorldProcess, ManyPassProcess):
    """Process class that adds its participants to the existing world, after it ends."""

    _passes = 1

    def _handle_participant(self, participant: IUpdatable) -> None:
        self.world.add_inhabitant(participant)


class UnitKillProcess(FocusedEvent, WorldProcess, ManyPassProcess):
    """
    Process class that removes its participants from the existing world, after
    it ends.
    """

    _passes = 1

    def _handle_participant(self, participant: IUpdatable) -> None:
        self.world.remove_inhabitant(participant)


class DelayedProcess(Process, ABC):
    """
    Process class that delays the execution of its logic for a certain number of
    updates.

    Number of updates is set by the _ticks_of_inactivity attribute.
    """

    _ticks_of_inactivity: int

    def activate_delay(self) -> None:
        """Logic execution delay resume method."""

        self.state = SleepProcessState(self, self._ticks_of_inactivity)


class CustomBilateralProcessFactory(IBilateralProcessFactory, ABC):
    """BilateralProcessFactory interface implementation class."""

    def __init__(self, process_type: type):
        self._process_type = process_type

    @property
    def process_type(self) -> type:
        return self._process_type

    def __call__(self, active_unit: IUpdatable, passive_unit: IUpdatable) -> Process:
        return self.process_type(active_unit, passive_unit)


class IProcessKeeper(ABC):
    """Class interface that implements logic as input processes and manages them."""

    @property
    @abstractmethod
    def processes(self) -> frozenset[IProcess]:
        """Active processes property."""

    @property
    @abstractmethod
    def completed_processes(self) -> frozenset[IProcess]:
        """Property of processes that have completed their work."""

    @abstractmethod
    def add_process(self, process: IProcess) -> None:
        """Method for adding a new process."""

    @abstractmethod
    def remove_process(self, process: IProcess) -> None:
        """Method for forcibly removing an existing process."""

    def is_support_process(self, process: IProcess) -> Report:
        """Method for determining input process support as internal."""

    @abstractmethod
    def activate_processes(self) -> None:
        """Process management method."""

    @abstractmethod
    def clear_completed_processes(self) -> None:
        """Method for cleaning up completed processes."""


class ProcessKeeper(IProcessKeeper, ABC):
    """ProcessKeeper interface implementation class."""

    _process_adding_report_analyzer = ReportAnalyzer((BadReportHandler(
        UnsupportedProcessError,
        "Process keeper unsupported process"
    ), ))

    def __init__(self):
        self._processes = set()
        self.__completed_processes = list()

    @property
    def processes(self) -> frozenset[IProcess]:
        return frozenset(self._processes)

    @property
    def completed_processes(self) -> frozenset[IProcess]:
        return frozenset(self.__completed_processes)

    def is_support_process(self, process: IProcess) -> Report:
        return Report(isinstance(process, IProcess))

    def add_process(self, process: IProcess) -> None:
        self._process_adding_report_analyzer(self.is_support_process(process))
        self._processes.add(process)

    def remove_process(self, process: IProcess) -> None:
        self._processes.remove(process)

    def activate_processes(self) -> None:
        processes_to_update, self._processes = self._processes, set()

        for process in processes_to_update:
            if type(process.state) is CompletedProcessState:
                self.__completed_processes.append(process)
            else:
                self._processes.add(process)
                process.update()

    def clear_completed_processes(self) -> None:
        self.__completed_processes = list()


class MultitaskingUnit(ProcessKeeper, IUpdatable, ABC):
    """Unit class implementing process support."""


class InteractiveMixin(IInteractive, ABC):
    """Interactive base implementation class."""

    _interaction_report_analyzer = ReportAnalyzer((BadReportHandler(
        InteractionError,
        "Object can't interact"
    ), ))

    def interact_with(self, passive: object) -> None:
        self._interaction_report_analyzer(self.is_support_interaction_with(passive))
        self._handle_interaction_with(passive)

    @abstractmethod
    def _handle_interaction_with(self, passive: object) -> None:
        """Method that implements the logic of interaction with a particular unit."""


class _ObjectFactoriesCash(NamedTuple):
    """Factory cache storage structure for a unit."""

    object_: object
    factories: tuple[IBilateralProcessFactory | StrictToParticipantsProcess]


class ProcessInteractiveMixin(InteractiveMixin, ProcessKeeper, ABC):
    """
    Mixin that implements interaction by creating two-way processes by object
    type.

    The two-way process factories for interaction are stored in the
    _bilateral_process_factories attribute. The content of the attribute can
    represent the factories of the corresponding processes, or process types
    strictly related to the state of the participants.
    """

    _bilateral_process_factories: Iterable[IBilateralProcessFactory | type]

    def is_support_interaction_with(self, passive: object) -> Report:
        return (
            Report(
                bool(self._get_suported_process_factories_for(passive)),
                error=IncorrectUnitInteractionError("No possible processes to occur")
            )
        )

    def _handle_interaction_with(self, passive: object) -> None:
        for factory in self._get_suported_process_factories_for(passive):
            process = factory(self, passive)
            process.start()

            self.add_process(process)

    def _get_suported_process_factories_for(self, passive: object) -> tuple[IBilateralProcessFactory]:
        """Method for getting corresponding factories by object."""

        return self.__get_cachedly_suported_process_factories_for(passive)

    def __get_cachedly_suported_process_factories_for(self, passive: object) -> tuple[IBilateralProcessFactory]:
        """Method for getting matching factories by unit using cache."""

        if passive is self.__cashed_factories_for_object.object_:
            return self.__cashed_factories_for_object.factories

        factories = tuple(
            (
                factory.process_type if hasattr(factory, 'process_type') else factory
            ).is_support_participants((self, passive))
            for factory in self._bilateral_process_factories
        )
        self.__cashed_factories_for_object = _ObjectFactoriesCash(passive, factories)

        return factories

    __cashed_factories_for_object: _ObjectFactoriesCash = _ObjectFactoriesCash(object(), tuple())


class InteractiveUnit(InteractiveMixin, IUpdatable, ABC):
    """Unit class that can interact with other units."""


class Dependent:
    """Class annotating the need for a dependency on another object."""

    master: Optional[object] = None


class StrictDependent(Dependent, StrictToStateMixin, StylizedMixin):
    """Dependent child class that must have a master for the correct state."""

    _repr_fields = (Field("master"), )
    _state_report_analyzer = ReportAnalyzer((BadReportHandler(UnmetDependencyError), ))

    def _is_correct(self) -> Report:
        return Report(
            self.master is not None,
            error=UnmetDependencyError(f"{self} must have a master")
        )


class StructuredPartDiscreteMixin(IDiscretable, ABC, metaclass=AttributesTransmitterMeta):
    """
    Class that allows you to structure attributes that have parts of an object.

    The names of the attributes that store the parts of an object are in the
    _part_attribute_names attribute.
    """

    _attribute_names_to_parse = '_part_attribute_names',
    _part_attribute_names: tuple[str]

    @property
    def parts(self) -> frozenset[object]:
        parts = self._get_parts()

        for part in parts:
            if isinstance(part, Dependent):
                part.master = self

        return parts

    def _get_parts(self) -> frozenset[object]:
        parts = list()

        for part_attribute_name in self._part_attribute_names:
            if not hasattr(self, part_attribute_name):
                continue

            attribute_value = getattr(self, part_attribute_name)

            append_method = getattr(
                parts,
                'extend' if isinstance(attribute_value, Iterable) else 'append'
            )

            append_method(attribute_value)

        return frozenset(parts)


class DeepPartDiscreteMixin(IDiscretable, ABC):
    """Mixin with the implementation of getting all parts for the Discrete interface."""

    @property
    def deep_parts(self) -> frozenset[object]:
        found_parts = set()

        for part in self.parts:
            found_parts.add(part)

            if hasattr(part, "deep_parts"):
                found_parts.update(part.deep_parts)
            elif hasattr(part, "parts"):
                found_parts.update(part.parts)

        return found_parts


class DiscreteUnit(IUpdatable, StructuredPartDiscreteMixin, DeepPartDiscreteMixin, ABC):
    """Discrete unit class containing other units."""

    @abstractmethod
    def init_parts(self, *args, **kwargs) -> None:
        """Method for initializing parts of a unit."""


class DiscreteUnitFactory(ABC):
    """
    Factory class for simultaneous creation and initialization of parts of a
    discrete unit.

    Creates a unit from the input arguments and initializes the parts using the
    stored.

    Stores the unit's factory in the _unit_factory attribute and the arguments
    to initialize its parts in the _unit_part_initialization_arguments attribute.
    """

    _unit_factory: Callable[[], DiscreteUnit]
    _unit_part_initialization_arguments: Arguments

    def __call__(self, *args, **kwargs) -> DiscreteUnit:
        unit = self._unit_factory(*args, **kwargs)
        self._unit_part_initialization_arguments.call_for(unit.init_parts)

        return unit


class CustomDiscreteUnitFactory(DiscreteUnitFactory):
    """DiscreteUnitFactory with input attributes to create a unit and its parts."""

    def __init__(self, unit_factory: Callable[[], DiscreteUnit], unit_part_initialization_arguments: Arguments):
        self._unit_factory = unit_factory
        self._unit_part_initialization_arguments = unit_part_initialization_arguments


class AvatarKeeper(IAvatarKeeper, ABC):
    """
    Implementation of the avatar keeper interface that initializes the avatar
    through the factory located in the _avatar_factory attribute.
    """

    _avatar_factory: IAvatarFactory

    def __init__(self):
        self._avatar = self._avatar_factory(self)

    @property
    def avatar(self) -> IAvatar:
        return self._avatar


class ZoneKeeper(ABC):
    """Class having a specific body as a zone."""

    _zone_factory: IZoneFactory

    def __init__(self):
        self._zone = self._zone_factory(self)

    @property
    def zone(self) -> Figure:
        return self._zone


class PositionalKeeper(ZoneKeeper, IPositional, ABC):
    """Сlass that has a location point."""

    _zone_factory = CustomFactory(lambda unit: Site(unit.position))

    def __init__(self, position: Vector):
        self._position = position
        super().__init__()

    @property
    def position(self) -> Vector:
        return self._position


class StaticAvatarKeeper(PositionalKeeper, AvatarKeeper):
    """Avatar keeper child class having a statically assigned position."""

    def __init__(self, position: Vector):
        super().__init__(position)
        AvatarKeeper.__init__(self)


class MovablePositionalKeeper(PositionalKeeper, IMovable, ABC):
    """Сlass providing dynamic position."""

    def __init__(self, position: Vector):
        super().__init__(position)
        self.__previous_position = self.position

    @property
    def previous_position(self) -> Vector:
        """Property of the position the unit had before the start of the last move."""

        return self.__previous_position

    @property
    @abstractmethod
    def next_position(self) -> Vector:
        """Property that defines the next position when moving."""

    def move(self) -> None:
        self.__previous_position = self._position
        self._position = self.next_position

        self._update_zone_position()

    def _update_zone_position(self) -> None:
        """
        Method of movement of a object's zone according to the vector of the last
        movement of the object itself.
        """

        self._zone.move_by(DynamicTransporter(self.position - self.previous_position))


class ProcessMovablePositionalKeeper(MovablePositionalKeeper, ABC):
    """
    Movable class delegating calculation of next position to a special process.

    Creates a moving process by the corresponding _moving_process_factory attribute.
    """

    _moving_process_factory: Callable[[Self], 'IMovingProcess']

    def __init__(self, position: Vector):
        super().__init__(position)
        self._moving_process = self._moving_process_factory(self)

    @property
    def moving_process(self) -> 'IMovingProcess':
        return self._moving_process

    @property
    def next_position(self) -> Vector:
        return self._moving_process.next_subject_position

    def move(self) -> None:
        self._moving_process.update()
        super().move()

        self._moving_process.state = MovingProcessState(self._moving_process)


class IMovingProcess(IProcess, ABC):
    """Process interface that calculates the next position of an object."""

    @property
    @abstractmethod
    def subject(self) -> ProcessMovablePositionalKeeper:
        """Subject property for which the next position is calculated."""

    @property
    @abstractmethod
    def next_subject_position(self) -> Vector:
        """Subject's computed next position property."""


class MovingProcess(Process, IMovingProcess, ABC):
    """MovingProcess Interface Implementation."""

    is_support_participants = CallableProxyReporter((TypeReporter((ProcessMovablePositionalKeeper, )), ))

    def __init__(self, subject: ProcessMovablePositionalKeeper):
        self._subject = subject

    @property
    def participants(self) -> tuple[ProcessMovablePositionalKeeper]:
        return self._subject

    @property
    def subject(self) -> ProcessMovablePositionalKeeper:
        return self._subject


class ProxyMovingProcess(ProxyProcess, IMovingProcess, ABC):
    """Process proxy class for a mobile process."""

    @property
    def subject(self) -> ProcessMovablePositionalKeeper:
        return self.process.subject

    @property
    def next_subject_position(self) -> Vector:
        return self.process.next_subject_position


class SpeedLimitedProxyMovingProcess(ProxyMovingProcess):
    """Proxy moving process that limits the length of the motion vector."""

    def __init__(self, process: MovingProcess, speed_limit: int | float):
        super().__init__(process)
        self._speed_limit = speed_limit

    @property
    def speed_limit(self) -> int | float:
        """Motion vector length constraint property."""

        return self._speed_limit

    @property
    def next_subject_position(self) -> Vector:
        vector_to_next_position = (
            self.process.next_subject_position
            - self.process.subject.previous_position
        )

        return self.process.subject.position + (
            vector_to_next_position
            if vector_to_next_position.length <= self.speed_limit
            else vector_to_next_position.get_reduced_to_length(self.speed_limit)
        )


class MovingProcessState(FlagProcessState):
    """Flag of the moving process indicating the movement of a movable object."""


class DirectedMovingProcess(MovingProcess):
    """Moving process class using a public vector."""

    def __init__(self, subject: ProcessMovablePositionalKeeper):
        super().__init__(subject)
        self.vector_to_next_subject_position = Vector()

    @property
    def next_subject_position(self) -> Vector:
        return self.subject.position + self.vector_to_next_subject_position

    def _handle(self):
        pass


class ImpulseMovingProcess(DirectedMovingProcess, ABC):
    """
    Directed Moving Process class that changes the motion vector after the
    movement has been made.
    """

    _impulse_changer: IPointChanger

    def _handle(self):
        if isinstance(self.state, MovingProcessState):
            self.vector_to_next_point = self._impulse_changer(self.vector_to_next_point)


class AbruptImpulseProcess(ImpulseMovingProcess):
    """ImpulseMovingProcess class that resets the motion vector."""

    _impulse_changer = CustomFactory(lambda original_vector: Vector())


class MultilayerProcessMovablePositionalKeeperMeta(AttributesTransmitterMeta):
    """
    Meta class for combining all factories of proxy moving processes into one.

    Merges factories from the _proxy_moving_process_factories attribute into the
    _multilayer_proxy_process_factory attribute.
    """

    _attribute_names_to_parse = ('_proxy_moving_process_factories', )

    _proxy_moving_process_factories: Iterable[Callable[[IProcess], IProcess]]
    _multilayer_proxy_process_factory: Optional[Callable[[Self], IMovingProcess]] = None

    def __new__(cls, class_name: str, super_classes: tuple, attributes: dict):
        isinstance_type = super().__new__(cls, class_name, super_classes, attributes)

        for proxy_moving_process_factory in isinstance_type._proxy_moving_process_factories:
            isinstance_type._multilayer_proxy_process_factory = CustomDecoratorFactory(
                proxy_moving_process_factory,
                (
                    isinstance_type._multilayer_proxy_process_factory
                    if isinstance_type._multilayer_proxy_process_factory is not None
                    else isinstance_type._proxy_moving_process_factories[0]
                )
            )

        return isinstance_type


class MultilayerProcessMovablePositionalKeeper(ProcessMovablePositionalKeeper, ABC, metaclass=MultilayerProcessMovablePositionalKeeperMeta):
    """
    Child class ProcessMovablePositionalKeeper with native implementation of
    building a motion process factory by the proxy process factories.
    """

    def __init__(self, position: Vector):
        self._moving_process_factory = CustomDecoratorFactory(
            self._multilayer_proxy_process_factory,
            self._moving_process_factory
        ) if self._multilayer_proxy_process_factory else self._moving_process_factory

        super().__init__(position)


class MultilayerProcessMovableAvatarKeeper(MultilayerProcessMovablePositionalKeeper, AvatarKeeper):
    """
    AvatarKeeper child class that implements positioning by inheriting from
    MultilayerProcessMovablePositionalKeeper.
    """

    def __init__(self, position: Vector):
        super().__init__(position)
        AvatarKeeper.__init__(self)


class WorldInhabitantsHandler(ABC):
    """Class that handles objects in a world."""

    _inhabitant_suitabing_report_analyzer = ReportAnalyzer(
        (BadReportHandler(UnsupportedInhabitantForHandlerError), )
    )

    def __init__(self, world: 'World'):
        self.world = world

    def __call__(self, inhabitants: Iterable) -> None:
        for inhabitant in inhabitants:
            self._inhabitant_suitabing_report_analyzer(self.is_inhabitant_suitable(inhabitant))

        self._handle_inhabitants(inhabitants)

    @abstractmethod
    def is_inhabitant_suitable(self, inhabitant: object) -> Report:
        """Method that returns a inhabitant support report for handling."""

    @abstractmethod
    def _handle_inhabitants(self, inhabitants: Iterable[IUpdatable]) -> None:
        """Handling method of world's inhabitants."""


class UnscrupulousWorldInhabitantsHandler(WorldInhabitantsHandler, ABC):
    """WorldInhabitantsHandler child class that handles each inhabitant."""

    def is_inhabitant_suitable(self, inhabitant: object) -> Report:
        return Report(True)


class TypeSuportingWorldInhabitantsHandler(WorldInhabitantsHandler, ABC, metaclass=TypeReporterKeeperMeta):
    """
    WorldInhabitantsHandler class implementing inhabitant support by delegating
    a type reporter.
    """

    def is_inhabitant_suitable(self, inhabitant: object) -> Report:
        return self._type_reporter.create_report_of((inhabitant, ))


class FocusedWorldInhabitantsHandler(WorldInhabitantsHandler, ABC):
    """WorldInhabitantsHandler child class uniformly handles inhabitants."""

    def _handle_inhabitants(self, inhabitants: Iterable) -> None:
        for inhabitant in inhabitants:
            self._handle_inhabitant(inhabitant)

    @abstractmethod
    def _handle_inhabitant(self, inhabitant: object) -> None:
        """Single inhabitant handling method."""


class InhabitantUpdater(FocusedWorldInhabitantsHandler, TypeSuportingWorldInhabitantsHandler):
    """WorldInhabitantsHandler child class updating inhabitants."""

    _suported_types = (IUpdatable, )

    def _handle_inhabitant(self, inhabitant: IUpdatable) -> None:
        inhabitant.update()


class InhabitantProcessesActivator(FocusedWorldInhabitantsHandler, TypeSuportingWorldInhabitantsHandler):
    """WorldInhabitantsHandler child class activating processes inside process keepers."""

    _suported_types = (IProcessKeeper, )

    def _handle_inhabitant(self, inhabitant: object) -> None:
        inhabitant.clear_completed_processes()
        inhabitant.activate_processes()


class WorldProcessesActivator(ProcessKeeper, FocusedWorldInhabitantsHandler, TypeSuportingWorldInhabitantsHandler):
    """
    WorldInhabitantsHandler child class connecting world processes from
    inhabitants with the world.
    """

    _suported_types = (IProcessKeeper, )

    def __init__(self, world: 'World'):
        super().__init__()
        super(TypeSuportingWorldInhabitantsHandler, self).__init__(world)

    def add_process(self, process: WorldProcess) -> None:
        process.world = self.world
        super().add_process(process)

    def _handle_inhabitants(self, inhabitants: Iterable[WorldProcess]) -> None:
        self.clear_completed_processes()
        super()._handle_inhabitants(inhabitants)
        self.activate_processes()

    def _handle_inhabitant(self, inhabitant: WorldProcess) -> None:
        for process in inhabitant.processes:
            if isinstance(process, WorldProcess):
                inhabitant.remove_process(process)
                self.add_process(process)


class RenderResourceParser(WorldInhabitantsHandler, IRenderRersourceKeeper, ABC):
    """
    RenderRersourceKeeper class that takes its render resource packs thanks to
    handling the inhabitants of the world.
    """

    def __init__(self, world: 'World'):
        super().__init__(world)
        self._parsed_resource_packs = list()

    @property
    def render_resource_packs(self) -> tuple[ResourcePack]:
        return tuple(self._parsed_resource_packs)

    def clear_parsed_resource_packs(self) -> None:
        self._parsed_resource_packs = list()

    def _handle_inhabitants(self, inhabitants: Iterable) -> None:
        self.clear_parsed_resource_packs()
        super()._handle_inhabitants(inhabitants)


class InhabitantAvatarRenderResourceParser(RenderResourceParser, FocusedWorldInhabitantsHandler, TypeSuportingWorldInhabitantsHandler):
    """RenderResourceParser taking packs from avatars of avatar keeper inhabitants."""

    _suported_types = (AvatarKeeper, )

    def _handle_inhabitant(self, inhabitant: AvatarKeeper) -> None:
        inhabitant.avatar.update()
        self._parsed_resource_packs.extend(inhabitant.avatar.render_resource_packs)


class AvatarRenderResourceParser(RenderResourceParser, FocusedWorldInhabitantsHandler, TypeSuportingWorldInhabitantsHandler):
    """RenderResourceParser taking packs from avatars inhabited in the world."""

    _suported_types = (IAvatar, )

    def _handle_inhabitant(self, inhabitant: IAvatar) -> None:
        self._parsed_resource_packs.extend(inhabitant.render_resource_packs)


class RelationsActivator(WorldInhabitantsHandler):
    """
    WorldInhabitantsHandler activating the relations of the objects inhabited in
    the world.
    """

    def _handle_inhabitants(self, inhabitants: Iterable) -> None:
        for active_inhabitants in inhabitants:
            if not isinstance(active_inhabitants, IInteractive):
                continue

            passive_inhabitants = set(inhabitants)
            passive_inhabitants.remove(active_inhabitants)

            for passive_inhabitant in passive_inhabitants:
                active_inhabitants.interact_with(passive_inhabitant)


class InhabitantMover(FocusedWorldInhabitantsHandler, TypeSuportingWorldInhabitantsHandler):
    """WorldInhabitantsHandler activating movement of moving inhabitants."""

    _suported_types = (IMovable, )

    def _handle_inhabitant(self, inhabitant: IMovable) -> None:
        inhabitant.move()


class World(IUpdatable, DeepPartDiscreteMixin, ABC):
    """
    The domain object habitat class.

    Delegates processing of domain entities to special handlers.
    Creates handlers using factories stored in _inhabitant_handler_factories
    attribute.
    """

    _inhabitant_handler_factories: Iterable[Callable[[Self], WorldInhabitantsHandler]]

    def __init__(self, inhabitants: Iterable = tuple()):
        self.__inhabitant = set()
        self._inhabitant_handlers = tuple(
            inhabitant_handler_factory(self)
            for inhabitant_handler_factory in self._inhabitant_handler_factories
        )

        for inhabitant in inhabitants:
            self.add_inhabitant(inhabitant)

    @property
    def parts(self) -> frozenset:
        return frozenset(self.__inhabitant)

    @property
    def inhabitant_handlers(self) -> tuple[WorldInhabitantsHandler]:
        """Property of handlers for world objects."""

        return self._inhabitant_handlers

    def is_inhabited_for(self, inhabitant: object) -> Report:
        return Report(not isinstance(inhabitant, World))

    def add_inhabitant(self, inhabitant: IUpdatable) -> None:
        if not self.is_inhabited_for(inhabitant):
            raise NotSupportPartError(f"World {self} does not support {inhabitant}")

        self.__inhabitant.add(inhabitant)

    def remove_inhabitant(self, inhabitant: IUpdatable) -> None:
        self.__inhabitant.remove(inhabitant)

    def update(self) -> None:
        for inhabitant_handler in self._inhabitant_handlers:
            inhabitant_handler(
                tuple(
                    inhabitant for inhabitant in self.deep_parts
                    if inhabitant_handler.is_inhabitant_suitable(inhabitant)
                )
            )


class CustomWorld(World):
    """World class using input handler factories."""

    def __init__(
        self,
        inhabitants: Iterable = tuple(),
        inhabitant_handler_factories: Iterable[Callable[[World], WorldInhabitantsHandler]] = tuple()
    ):
        self._inhabitant_handler_factories = tuple(inhabitant_handler_factories)
        super().__init__(inhabitants)


class AppFactory(IAppFactory, metaclass=AttributesTransmitterMeta):
    """Class that implements the application factory interface."""

    _attribute_names_to_parse = ('_loop_handler_factories', )

    _loop_handler_factories: Iterable[LoopHandler] = tuple()
    _updater_loop_handler_factory: LoopHandler = UpdaterLoopHandler

    _loop_factory: Callable[[Iterable[UpdaterLoopHandler]], ILoop] = CustomHandlerLoop
    _render_activator_factory: IRenderActivatorFactory = RenderActivator

    def __call__(
        self,
        world: World,
        renders: Iterable[IRender]
    ) -> ILoop:
        render_activator = self._render_activator_factory(
            self._get_resource_parsers_from(world),
            renders
        )

        return self._loop_factory((
            CustomFactory(self._updater_loop_handler_factory, (world, render_activator)),
            *self._loop_handler_factories
        ))

    def _get_resource_parsers_from(self, world: World) -> tuple[RenderResourceParser]:
        resource_parsers = tuple(filter(
            lambda handler: isinstance(handler, IRenderRersourceKeeper),
            world.inhabitant_handlers
        ))

        if resource_parsers:
            return resource_parsers

        raise InvalidWorldError(f"World {world} does not have resource parsers for render")


class CustomAppFactory(AppFactory):
    """AppFactory class with input factories."""

    def __init__(
        self,
        loop_handler_factories: Iterable[LoopHandler] = tuple(),
        loop_factory: Callable[[Iterable[UpdaterLoopHandler]], ILoop] = CustomHandlerLoop,
        updater_loop_handler_factory: LoopHandler = UpdaterLoopHandler,
        render_activator_factory: IRenderActivatorFactory = RenderActivator
    ):
        self._loop_factory = loop_factory
        self._render_activator_factory = render_activator_factory
        self._loop_handler_factories = loop_handler_factories
        self._updater_loop_handler_factory = updater_loop_handler_factory



class SimulationError(Exception):
    pass


class ProcessKeeperError(SimulationError):
    pass


class UnsupportedProcessError(ProcessKeeperError):
    pass


class InteractionError(SimulationError):
    pass


class UnmetDependencyError(SimulationError):
    pass


class WorldInhabitantsHandlerError(SimulationError):
    pass


class UnsupportedInhabitantForHandlerError(WorldInhabitantsHandlerError):
    pass


class UnitError(SimulationError):
    pass


class DiscreteError(SimulationError):
    pass


class NotSupportPartError(DiscreteError):
    pass


class ProcessError(SimulationError):
    pass


class ProcessAlreadyCompletedError(ProcessError):
    pass


class ProcessHasNotStartedError(ProcessError):
    pass


class ProcessStateError(ProcessError):
    pass


class ProcessStateIsNotValidError(ProcessStateError):
    pass


class ProcessIsNoLongerSleepingError(ProcessStateIsNotValidError):
    pass


class WorldProcessError(ProcessError):
    pass


class AppFactoryError(SimulationError):
    pass


class InvalidWorldError(AppFactoryError):
    pass



from abc import ABC, abstractmethod
from typing import Iterable, NewType, Optional, Callable, Self

from pygame import *

from sim32.pygame_integration.errors import PygameEventHandlerError
from sim32.pygame_integration.resources import *
from sim32.interfaces import IUpdatable
from sim32.renders import Render, SurfaceKeeper, TypedResourceHandler, ResourcePack, resource_handler
from sim32.geometry import Vector
from sim32.tools import *


class PygameSurfaceRender(SurfaceKeeper, Render):
    """
    Pygame surface render class.

    Uses pygame render function attribute representations as a function definer
    with appropriate attributes.
    """

    _resource_handler_wrapper_factory = TypedResourceHandler

    def __init__(self, surfaces: Iterable[Surface], background_color: RGBAColor = RGBAColor()):
        super().__init__(surfaces)
        self.background_color = background_color

    def _clear_surface(self, surface: any) -> None:
        surface.fill(tuple(self.background_color))

    @resource_handler(Surface)
    def _handle_pygame_surface(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        surface.blit(resource_pack.resource, resource_pack.point.coordinates)

    @resource_handler(PygamePolygon)
    def _handle_pygame_polygon(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        draw.polygon(
            surface,
            tuple(resource_pack.resource.color),
            tuple(
                (resource_pack.point + vector_to_point).coordinates
                for vector_to_point in resource_pack.resource.points
            ),
            resource_pack.resource.border_width
        )

    @resource_handler(PygameLine)
    def _handle_pygame_line(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        (draw.line if not resource_pack.resource.is_smooth else draw.aaline)(
            surface,
            tuple(resource_pack.resource.color),
            (resource_pack.resource.start_point + resource_pack.point).coordinates,
            (resource_pack.resource.end_point + resource_pack.point).coordinates,
            resource_pack.resource.border_width
        )

    @resource_handler(PygameLines)
    def _handle_pygame_lines(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        (draw.lines if not resource_pack.resource.is_smooth else draw.aalines)(
            surface,
            tuple(resource_pack.resource.color),
            resource_pack.resource.is_closed,
            tuple(
                (line_point + resource_pack.point).coordinates
                for line_point in resource_pack.resource.points
            ),
            resource_pack.resource.border_width
        )

    @resource_handler(PygameCircle)
    def _handle_pygame_circle(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        draw.circle(
            surface,
            tuple(resource_pack.resource.color),
            resource_pack.point.coordinates,
            resource_pack.resource.radius,
            resource_pack.resource.border_width
        )

    @resource_handler(PygameRectangle)
    def _handle_pygame_rect(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        draw.rect(
            surface,
            tuple(resource_pack.resource.color),
            (
                *resource_pack.point.coordinates,
                resource_pack.resource.width,
                resource_pack.resource.height
            ),
            resource_pack.resource.border_width
        )

    @resource_handler(PygameEllipse)
    def _handle_pygame_ellipse(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        draw.ellipse(
            surface,
            tuple(resource_pack.resource.color),
            (
                *resource_pack.point.coordinates,
                resource_pack.resource.width,
                resource_pack.resource.height
            ),
            resource_pack.resource.border_width
        )

    @resource_handler(PygameArc)
    def _handle_pygame_arc(resource_pack: ResourcePack, surface: Surface, render: Self) -> None:
        draw.arc(
            surface,
            tuple(resource_pack.resource.color),
            (
                *resource_pack.point.coordinates,
                resource_pack.resource.width,
                resource_pack.resource.height
            ),
            resource_pack.resource.start_angle,
            resource_pack.resource.stop_angle,
            resource_pack.resource.border_width
        )


PygameEvent: NewType = object


class IPygameEventHandler(ABC):
    """Pygame event handler interface."""

    @abstractmethod
    def __call__(self, event: PygameEvent, controller: 'PygameEventController') -> None:
        """Handling logic start method."""

    @abstractmethod
    def is_support_handling_for(self, event: PygameEvent, controller: 'PygameEventController') -> bool:
        """Method for reporting a possible handling call."""


class PygameEventHandler(IPygameEventHandler, ABC):
    """Implementing the PygameEventHandler Interface."""

    def __call__(self, event: PygameEvent, controller: 'PygameEventController') -> None:
        if not self.is_support_handling_for(event, controller):
            raise PygameEventHandlerError(
                f"Event handler {self} doesn't support handling event {event} in controller {controller}"
            )

        self._handle(event, controller)

    @abstractmethod
    def _handle(self, event: PygameEvent, controller: 'PygameEventController') -> None:
        """Event handling logic method."""


class PygameEventHandlerWrapper(PygameEventHandler):
    """
    Pygame Event Handler class allows you to operate multiple handlers as a
    single object.
    """

    def __init__(self, handlers: Iterable[IPygameEventHandler]):
        self.handlers = tuple(handlers)

    def _handle(self, event: PygameEvent, controller: 'PygameEventController') -> None:
        for handler in self.handlers:
            if handler.is_support_handling_for(event, controller):
                handler(event, controller)


class EventSupportStackHandler(IPygameEventHandler, ABC):
    """
    Implementation of the PygameEventHandler interface with automatic support for
    determining support for event handling by configured attributes.
    """

    _support_event_types: Iterable
    _support_keys: Optional[Iterable] = None
    _support_buttons: Optional[Iterable] = None
    _is_strict: bool = True

    def is_support_handling_for(self, event: PygameEvent, controller: 'PygameEventController') -> bool:
        return (all if self._is_strict else any)((
            (event.key in self._support_keys) if hasattr(event, 'key') else self._support_keys is None,
            (event.button in self._support_buttons) if hasattr(event, 'button') else self._support_buttons is None
        )) if event.type in self._support_event_types else False


class ExitEventHandler(PygameEventHandler, EventSupportStackHandler):
    """PygameEventHandler class that handles the exit event with appropriate logic."""

    _support_event_types = (QUIT, )

    def _handle(self, event: PygameEvent, controller: 'PygameEventController') -> None:
        exit()


class IPygameEventGetter(ABC):
    """Pygame keeper interface."""

    @abstractmethod
    def get(self) -> Iterable[PygameEvent]:
        """Method to get events."""


class PygameEventController(LoopHandler):
    """
    LoopHandler class delegating the handling of pygame events.

    Gets events using the event getter and delegates them to input event handlers.
    """

    _event_getter: IPygameEventGetter

    def __init__(self, loop: HandlerLoop, handlers: Iterable[PygameEventHandler]):
        super().__init__(loop)
        self.handlers = tuple(handlers)

    def update(self) -> None:
        for event_ in self._event_getter.get():
            self._handle_event(event_)

    def _handle_event(self, event: PygameEvent) -> None:
        for event_handler in self.handlers:
            if event_handler.is_support_handling_for(event, self):
                event_handler(event, self)


class SyncPygameEventController(PygameEventController):
    """Pygame Event Controller class using the standard pygame event store."""

    _event_getter = event


class PygameDisplayUpdater(LoopHandler):
    """LoopHandler class that updates the main window created by pygame."""

    def update(self) -> None:
        display.flip()


class PygameClockSleepLoopHandler(TicksSleepLoopHandler, AlwaysReadyForSleepLoopHandler):
    """TicksSleepLoopHandler class that implements waiting through pygame clock."""

    _clock_factory: Callable[[Self], time.Clock] = CustomFactory(
        lambda pygame_sleep_handler: time.Clock()
    )

    def __init__(self, loop: HandlerLoop, ticks_to_sleep: int | float):
        super().__init__(loop, ticks_to_sleep)
        self._pygame_clock = self._clock_factory(self)

    def _sleep_function(self, ticks_to_sleep: int | float) -> None:
        self._pygame_clock.tick(self.ticks_to_sleep)



from sim32.errors.tool_errors import UnableToDivideError


class GeometryError(Exception):
    pass


class AxisPlaneDegreesError(GeometryError):
    pass


class VectorError(GeometryError):
    pass


class LineError(GeometryError):
    pass


class UnableToDivideVectorIntoPointsError(GeometryError, UnableToDivideError):
    pass


class FigureError(GeometryError):
    pass


class FigureIsNotCorrect(FigureError):
    pass


class FigureIsNotClosedError(FigureIsNotCorrect):
    pass



from pygame import *
from random import randint, choice

from sim32.avatars import ResourceAvatar, PrimitiveAvatar
from sim32.core import *
from sim32.tools import Timer
from sim32.pygame_integration import *


class MainHeroManagement(PygameEventHandler, EventSupportStackHandler):
    _right_movement_keys = (K_RIGHT, K_d)
    _left_movement_keys = (K_LEFT, K_a)
    _up_movement_keys = (K_UP, K_w)
    _down_movement_keys = (K_DOWN, K_s)

    _support_keys = (
        *_right_movement_keys,
        *_left_movement_keys,
        *_up_movement_keys,
        *_down_movement_keys
    )
    _support_event_types = (KEYDOWN, )

    def __init__(self, main_hero: InfinitelyImpulseUnit):
        self.main_hero = main_hero

    def _handle(self, event: PygameEvent, loop: HandlerLoop) -> None:
        impulse = Vector((0, 0))

        if event.key in self._right_movement_keys:
            impulse += Vector((self.main_hero._speed_limit, 0))
        if event.key in self._left_movement_keys:
            impulse -= Vector((self.main_hero._speed_limit, 0))

        if event.key in self._up_movement_keys:
            impulse -= Vector((0, self.main_hero._speed_limit))
        if event.key in self._down_movement_keys:
            impulse += Vector((0, self.main_hero._speed_limit))

        self.main_hero.moving_process.original_process.vector_to_next_subject_position = impulse


class TestObject(MultilayerProcessMovableAvatarKeeper):
    _moving_process_factory = AbruptImpulseProcess
    _proxy_moving_process_factories = (CustomFactory(SpeedLimitedProxyMovingProcess, 2), )

    _avatar_factory = CustomFactory(
        lambda unit: PrimitiveAvatar(
            unit,
            Circle(
                RGBAColor(choice(range(255)), choice(range(255)), choice(range(255))),
                choice(range(3, 50))
            )
        )
    )


class ObserveAvatar(ResourceAvatar):
    _resource_factory = CustomFactory(lambda _: Circle(RGBAColor(255, 0, 50), 0))

    def update(self) -> None:
        super().update()
        vector_to_observed = self.domain.position - self.domain.observed.position
        self.render_resource.radius = vector_to_observed.length


class ObserveUnit(MultilayerProcessMovableAvatarKeeper, IUpdatable):
    _avatar_factory = ObserveAvatar

    _moving_process_factory = AbruptImpulseProcess
    _proxy_moving_process_factories = (CustomFactory(SpeedLimitedProxyMovingProcess, 1), )

    def __init__(self, position: Vector, observed: IPositional):
        super().__init__(position)
        self.observed = observed

    def update(self) -> None:
        self._moving_process.original_process.vector_to_next_subject_position = (
            self.observed.position
            - self.position
        )


class SpawnerUnit(StaticAvatarKeeper, MultitaskingUnit):
    _avatar_factory = CustomFactory(lambda unit: PrimitiveAvatar(unit, None))

    def __init__(
        self,
        position: Vector,
        factory: Callable[[Vector], IPositional],
        spawn_zone: Iterable[Diapason],
        timer: Timer
    ):
        super().__init__(position)
        super(MultitaskingUnit, self).__init__()

        self.factory = factory
        self.spawn_zone = tuple(spawn_zone)
        self.timer = timer

    def update(self) -> None:
        if self.timer.is_time_over():
            generate_point = Vector(tuple(
                randint(coordinate_diapason.start, coordinate_diapason.end)
                for coordinate_diapason in self.spawn_zone
            ))

            self.add_process(
                UnitSpawnProcess((self.factory(generate_point), ))
            )
            self.timer.start()


black_unit = TestObject(Vector((200, 240)))
black_unit.avatar.render_resource = Circle(RGBAColor(), 20)

red_unit = ObserveUnit(Vector((100, 240)), black_unit)

unit_spawner = SpawnerUnit(
    Vector((320, 240)),
    CustomFactory(TestObject),
    (Diapason(640), Diapason(480)),
    Timer(3)
)

unit_spawner.avatar.render_resource = Circle(RGBAColor(red=255, green=243), 30)


CustomAppFactory((
    CustomFactory(
        SyncPygameEventController, (
            ExitEventHandler(),
            MainHeroManagement(black_unit)
        )
    ),
    PygameDisplayUpdater,
    CustomFactory(PygameClockSleepLoopHandler, 60)
))(
    CustomWorld(
        [black_unit, red_unit, unit_spawner],
        [InhabitantUpdater, WorldProcessesActivator, InhabitantMover, InhabitantAvatarRenderResourceParser]
    ),
    (
        PygameSurfaceRender(
            (display.set_mode((640, 480)), ),
            RGBAColor(232, 232, 232)
        ),
    )
).run()



from abc import ABC
from typing import Optional, Callable, Iterable

from colorama import init, Fore

from sim32.avatars import PrimitiveAvatar
from sim32.core import *
from sim32.renders import ConsoleCell, ConsoleRender
from sim32.geometry import Vector, AxisPlaneDegrees, DegreeMeasure
from sim32.tools import CustomFactory, Diapason, Arguments


init(autoreset=True)


class Snake(DiscreteUnit):
    _part_attribute_names = ('_head', '_tails')

    _head: Optional[MovablePositionalKeeper] = None
    _tails: Iterable = tuple()

    @property
    def head(self) -> Optional[MovablePositionalKeeper]:
        return self._head

    @property
    def tails(self) -> tuple[ProcessMovablePositionalKeeper]:
        return tuple(self._tails)

    def init_parts(
        self,
        head: MovablePositionalKeeper,
        tail_factory: Callable[[Vector], ProcessMovablePositionalKeeper],
    ) -> None:
        self._tails = list()
        self._head = head
        self._tail_factory = tail_factory

    def grow_tail(self, tail_length: int) -> None:
        for _ in range(tail_length):
            last_node = self._tails[-1] if self._tails else self._head

            self._tails.append(self._tail_factory(last_node.previous_position))

    def cut_tail(self, tail_length: int) -> None:
        self._tails = self._tails[:-tail_length]

    def update(self) -> None:
        for tail_index, tail in enumerate(self._tails):
            if not isinstance(tail.moving_process.original_process, DirectedMovingProcess):
                continue

            previous_unit = self.__get_prevous_node_by(tail_index)

            tail.moving_process.original_process.vector_to_next_subject_position = (
                previous_unit.position - tail.position
            )

    def __get_prevous_node_by(self, tail_index: int) -> MovablePositionalKeeper:
        return self._head if tail_index - 1 < 0 else self._tails[tail_index - 1]


class SnakeHead(MultilayerProcessMovableAvatarKeeper):
    _avatar_factory = CustomFactory(PrimitiveAvatar, ConsoleCell('#', Fore.LIGHTBLUE_EX))

    _moving_process_factory = AbruptImpulseProcess
    _proxy_moving_process_factories = (CustomFactory(SpeedLimitedProxyMovingProcess, 1), )


class SnakeTail(MultilayerProcessMovableAvatarKeeper):
    _avatar_factory = CustomFactory(PrimitiveAvatar, ConsoleCell('#', Fore.LIGHTWHITE_EX))

    _moving_process_factory = DirectedMovingProcess
    _proxy_moving_process_factories = (CustomFactory(SpeedLimitedProxyMovingProcess, 1), )


class SnakeEvent(Process, ABC):
    def __init__(self, snake: Snake):
        super().__init__()
        self.snake = snake

    @property
    def participants(self) -> tuple[Snake]:
        return (self.snake, )


class SnakeTailLengthKepeerEvent(SnakeEvent):
    def __init__(self, snake: Snake, tail_length_diapason: Diapason, tail_number: int = 1):
        super().__init__(snake)
        self.tail_number = tail_number
        self.tail_length_diapason = tail_length_diapason

    def _handle(self) -> None:
        if len(self.snake.tails) == self.tail_length_diapason.start:
            self.__is_growing_mode = True

        elif len(self.snake.tails) >= self.tail_length_diapason.end:
            self.__is_growing_mode = False

        getattr(self.snake, 'grow_tail' if self.__is_growing_mode else 'cut_tail')(self.tail_number)

    __is_growing_mode: bool = True


class SnakeHeadTurnEvent(SnakeEvent, DelayedProcess):
    _ticks_of_inactivity: int = 10

    def start(self) -> None:
        super().start()
        self._activate_snake_head_moving()

    def _handle(self) -> None:
        self._activate_snake_head_moving()
        self.__snake_head_vector = self.__snake_head_vector.get_rotated_by(
            AxisPlaneDegrees(0, 1, DegreeMeasure(90))
        )

        self.activate_delay()

    def _activate_snake_head_moving(self) -> None:
        self.snake.head.moving_process.original_process.vector_to_next_subject_position = self.__snake_head_vector

    __snake_head_vector = Vector((1, 0))


snake_factory = CustomDiscreteUnitFactory(
    Snake,
    Arguments.create_via_call(SnakeHead(Vector((10, 8))), SnakeTail)
)

snake = snake_factory()
snake.grow_tail(5)


CustomAppFactory([CustomFactory(StandardSleepLoopHandler, 0.5)])(
    CustomWorld(
        [SnakeHeadTurnEvent(snake), SnakeTailLengthKepeerEvent(snake, Diapason(2, 9)), snake],
        [InhabitantMover, InhabitantUpdater, InhabitantAvatarRenderResourceParser]
    ),
    (ConsoleRender(ConsoleCell('.')), )
).run()



from abc import ABC, abstractmethod
from typing import Iterable


class IUpdatable(ABC):
    """
    The interface of anything that can handle its state.

    Objects that are upcast to this interface and implemented with a focus on
    self-computation are called units.
    """

    @abstractmethod
    def update(self) -> None:
        """Main method for resuming | continuing computations."""


class IDiscretable(ABC):
    """Interface for objects delegating or processing any other objects."""

    @property
    @abstractmethod
    def parts(self) -> frozenset[object]:
        """The property of getting objects directly lying in this object."""

    @property
    def deep_parts(self) -> frozenset[object]:
        """Property to get all parts including parts of the parts themselves."""


class IInteractive(ABC):
    """Interface of an object capable of responding to another object."""

    @abstractmethod
    def interact_with(self, passive: object) -> None:
        """Method for starting interaction with the input object."""

    @abstractmethod
    def is_support_interaction_with(self, passive: object) -> 'Report':
        """Method describing object support for interacting with it."""


class IPositional(ABC):
    """Interface of an object located at some point."""

    @property
    @abstractmethod
    def position(self) -> 'Vector':
        """Object location property."""


class IMovable(ABC):
    """Interface for objects capable of moving."""

    @abstractmethod
    def move(self) -> None:
        """Direct Method of object motion."""


class IRenderRersourceKeeper(ABC):
    """Interface for objects containing data to be drawn."""

    @property
    @abstractmethod
    def render_resource_packs(self) -> tuple['ResourcePack']:
        """Property for packs of directly render resources."""


class IAvatar(IUpdatable, IRenderRersourceKeeper, ABC):
    """Interface for processing and directly storing processed render resources."""


class IAvatarKeeper(IPositional, ABC):
    """
    Interface that allows an avatar to build visual projections based on objects
    that implement this interface.
    """

    @property
    @abstractmethod
    def avatar(self) -> IAvatar:
        """Property of the object processing the visual representation."""


class ILoop(ABC):
    """
    Loop class representation interface for abstracting dependencies on how an
    loop starts.
    """

    @abstractmethod
    def run(self) -> None:
        """Loop Start Method to respectively start a loop."""

    @abstractmethod
    def finish(self) -> None:
        """Method for forcibly ending a loop."""


class IZone(ABC):
    """
    Interface for representing an object that occupies some space and the
    possibility of checking about the presence of other objects in this space.
    """

    @abstractmethod
    def move_by(self, point_changer: 'IPointChanger') -> None:
        """
        A method for structured zone modification by abstracting from the direct
        implementation of the modification.

        Uses a point changer as an atomic implementation of changes, which is
        the argument of this method.
        """

    @abstractmethod
    def is_vector_passes(self, vector: 'VirtualVector') -> bool:
        """Method for finding an input vector in a zone."""

    @abstractmethod
    def is_vector_entered(self, vector: 'VirtualVector') -> bool:
        """Method for finding an end point of an input vector in a zone."""

    @abstractmethod
    def is_point_inside(self, point: 'Vector') -> bool:
        """Method for finding an input in a zone."""


class IRenderActivatorFactory(ABC):
    """Factory interface for implementation of RenderActivatorFactory interface."""

    @abstractmethod
    def __call__(
        self,
        rersource_keepers: Iterable[IRenderRersourceKeeper],
        redners: Iterable['IRender']
    ) -> 'RenderActivator':
        """Method creating an object."""


class IAppFactory(ABC):
    """
    Factory interface for creating a loop that implements the connection between
    the domain and its rendering.

    Due to the specific form of such loops, they are further called applications,
    since they implement its launch.
    """

    def __call__(
        self,
        world: 'World',
        renders: Iterable['IRenderResourceParser']
    ) -> 'ILoop':
        """
        Method for creating a loop, where the domain is an input world instance
        and the rendering implementations are input renders.
        """


class IZoneFactory(ABC):
    """Zone factory interface for an input object."""

    @abstractmethod
    def __call__(self, unit: IUpdatable) -> 'IZone':
        """Method creating an object."""


class IBilateralProcessFactory(ABC):
    """
    Interface for creating a separate process that implements the relationship
    of two objects.
    """

    @property
    @abstractmethod
    def process_type(self) -> type:
        """Property holding the type of the subsequently created process."""

    @abstractmethod
    def __call__(self, active: object, passive: object) -> 'IProcess':
        """
        Process creation method that has two arguments: a process's exciter and
        its acceptor.
        """


class IAvatarFactory(ABC):
    """Factory interface for creating an avatar for an unit."""

    @abstractmethod
    def __call__(self, unit: 'PositionalUnit') -> IAvatar:
        """Method creating an object."""



class ToolError(Exception):
    pass


class DividerError(ToolError):
    pass


class UnableToDivideError(DividerError):
    pass


class RGBAColorError(ToolError):
    pass


class ColorCoordinateError(RGBAColorError):
    pass


class AlphaChannelError(RGBAColorError):
    pass


class TimerError(ToolError):
    pass



from setuptools import setup


PACKAGE_NAME = 'sim32'
VERSION = "2.0.0"

REQUIRES = ["beautiful_repr==1.1.1", "pyoverload==1.1.24", "colorama==0.4.6"]


with open('README.md') as readme_file:
    LONG_DESCRIPTION = readme_file.read()

setup(
    name=PACKAGE_NAME,
    description="Library to help you write games",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version=VERSION,
    url="https://github.com/TheArtur128/SimEngine",
    download_url=f"https://github.com/TheArtur128/SimEngine/archive/refs/tags/{VERSION}.zip",
    author="Arthur",
    author_email="s9339307190@gmail.com",
    install_requires=REQUIRES,
    python_requires='>=3.10',
    packages=[
        PACKAGE_NAME,
        f"{PACKAGE_NAME}.pygame_integration",
        f"{PACKAGE_NAME}.examples",
        f"{PACKAGE_NAME}.errors"
    ],
    package_dir={
        PACKAGE_NAME: "sim32",
        f"{PACKAGE_NAME}.pygame_integration": "pygame_integration",
        f"{PACKAGE_NAME}.examples": "examples",
        f"{PACKAGE_NAME}.errors": "errors"
    }
)



from abc import ABC, abstractmethod, ABCMeta
from dataclasses import dataclass
from time import sleep, time, ctime
from threading import Thread
from typing import Iterable, Callable, Self, Protocol, NamedTuple
from math import floor, copysign
from enum import IntEnum
from functools import wraps

from beautiful_repr import StylizedMixin, Field, TemplateFormatter

from sim32.interfaces import IUpdatable, ILoop
from sim32.errors.tool_errors import *


@dataclass(frozen=True)
class Arguments:
    """
    Dataclass representing storage for args and kwargs. Has two corresponding
    fields: args and kwargs.
    """

    args: tuple
    kwargs: dict

    def call_for(self, function: Callable) -> any:
        """Method for calling a function with arguments in a dataclass object."""

        return function(*self.args, **self.kwargs)

    @classmethod
    def create_via_call(cls, *args, **kwargs) -> Self:
        """
        Method that allows you to create Arguments with args and kwargs as
        inputs to this function.
        """

        return cls(args, kwargs)


class DecoratorFactory(ABC):
    """
    Class for the ability to construct a component and wrap it in a proxy at the
    time of creation.

    Descendants must have _decorator_factory proxy factory and _nested_factory
    factory of the generated component.
    """

    _decorator_factory: Callable[[Callable], any]
    _nested_factory: Callable

    def __call__(self, *args_for_nested_factory, **kwargs_for_nested_factory) -> any:
        """Method for directly creating and wrapping a component in a decorator."""

        return self._decorator_factory(
            self._nested_factory(
                *args_for_nested_factory,
                **kwargs_for_nested_factory
            )
        )


class CustomDecoratorFactory(DecoratorFactory):
    """
    DecoratorFactory takes _nested_factory and _decorator_factory as their
    respective input arguments on object initialization.
    """

    def __init__(self, decorator_factory: Callable, nested_factory: Callable):
        self._decorator_factory = decorator_factory
        self._nested_factory = nested_factory

    @property
    def decorator_factory(self) -> Callable[[Callable], any]:
        return self._decorator_factory

    @property
    def nested_factory(self) -> Callable:
        return self._nested_factory


class CustomArgumentFactory(ABC):
    """
    Factory class that is a layer for creating an object using an interface
    other than the direct method of initializing this object.

    Stores the arguments and pads them to the call to the internal factory that
    it delegates to create the desired object.

    factory - Internal factory to which object creation is delegated.
    is_stored_arguments_first - Boolean argument specifying the position of the
    stored arguments when delegating.
    """

    factory: Callable
    is_stored_arguments_first: bool = False

    def __init__(self, *args_for_factory, **kwargs_for_factory):
        """Specifies arguments for the internal factory."""

        self.arguments_for_factory = Arguments.create_via_call(
            *args_for_factory,
            **kwargs_for_factory
        )

    def __call__(self, *args, **kwargs) -> any:
        """Creates some object by internal and input arguments."""

        argument_groups = [args, self.arguments_for_factory.args]

        if self.is_stored_arguments_first:
            argument_groups.reverse()

        return self.factory(
            *argument_groups[0],
            *argument_groups[1],
            **kwargs,
            **self.arguments_for_factory.kwargs
        )


class CustomFactory(CustomArgumentFactory):
    """
    Specifies the attributes of the parent's behavior during initialization.
    """

    def __init__(
        self,
        factory: Callable,
        *args_for_factory,
        is_stored_arguments_first: bool = False,
        **kwargs_for_factory
    ):
        self.factory = factory
        self.is_stored_arguments_first = is_stored_arguments_first
        super().__init__(*args_for_factory, **kwargs_for_factory)

    @classmethod
    def create_by_arguments(
        cls,
        factory: Callable,
        arguments: Arguments,
        is_stored_arguments_first: bool = False
    ) -> Self:
        """
        Method that simplifies the creation of a CustomFactory if there are
        Arguments for this.
        """

        return cls(
            factory,
            *arguments.args,
            is_stored_arguments_first=is_stored_arguments_first,
            **arguments.kwargs
        )


def get_collection_with_reduced_nesting_level_by(
    nesting_level: int,
    collection: Iterable
) -> list:
    """
    Function that allows you to get a lower nested representation of the input
    collection by unpacking nested collections, by the input nesting_level.
    """

    is_reduced = False

    while not is_reduced and nesting_level > 0:
        new_collection = list()
        is_reduced = True

        for item in collection:
            if isinstance(item, Iterable):
                is_reduced = False
                new_collection.extend(item)
            else:
                new_collection.append(item)

        nesting_level -= 1
        collection = new_collection

    return collection


def convert_fps_to_seconds(seconds: int | float) -> float:
    """Function that converts seconds to fps."""

    return 1 / seconds


class IValueTransformer(ABC):
    """
    Interface representing an atomic value converter to reduce dependencies on
    implementation of changes.
    """

    @abstractmethod
    def __call__(self, attribute_keeper: object, original_value: any) -> any:
        """Method for changing a value."""


class ForwardableValueTransformer(IValueTransformer):
    """
    Formal implementation of the ValueTransformer interface that does not change
    the value in any way.

    It makes sense to use it as a default implementation.
    """

    def __call__(self, attribute_keeper: object, original_value: any) -> any:
        return original_value


class ChangerPack(NamedTuple):
    """
    Storage class for modifying any collection and its items.

    Has two corresponding fields for changers.
    """

    item_changer: IValueTransformer
    collection_changer: IValueTransformer


class AttributesTransmitterMeta(ABCMeta):
    """
    Metaclass for automatically extending collection attributes of its instances
    by collections of the descendants.

    For correct use by classes, these classes must have a mandatory
    _attribute_names_to_parse attribute that describes the attributes to be
    extended by descendants. By default, this attribute is also expanded on a par
    with attributes whose names are stored in it. The final form of
    _attribute_names_to_parse is a dictionary that stores the attribute names as
    keys and as value changers that change the values of the inherited
    corresponding attributes, but for convenience, you can specify it as a
    collection of attribute names, subsequently converted to the final form.

    A child class can specify a _default_value_transformer that modifies inherited
    attributes and applies to attributes that do not have an explicit
    value_transformer in _attribute_names_to_parse.
    """

    _attribute_names_to_parse: Iterable[str] | dict[str, IValueTransformer | None]
    _default_value_transformer: IValueTransformer = ForwardableValueTransformer()

    def __new__(cls, class_name: str, super_classes: tuple, attributes: dict):
        """Assembly method of the class in which the attributes are converted and extended."""

        isinstance_type = super().__new__(cls, class_name, super_classes, attributes)

        isinstance_type._update_attribute_names()

        for attribute_name_to_parse in isinstance_type._deeply_get_attribute_names().keys():
            setattr(
                isinstance_type,
                attribute_name_to_parse,
                isinstance_type._parse_collection_by_attribute_name_from(
                    attributes,
                    attribute_name_to_parse
                )
            )

        return isinstance_type

    def _update_attribute_names(cls) -> None:
        """
        Method that converts _attribute_names_to_parse to its final form if it
        hasn't already, and extends it with ancestor class names.
        """

        if not isinstance(cls._attribute_names_to_parse, dict):
            cls._attribute_names_to_parse = dict.fromkeys(cls._attribute_names_to_parse)

        cls._attribute_names_to_parse = dict(
            item for parent_type in (*cls.__bases__, cls)
            if hasattr(parent_type, '_attribute_names_to_parse')
            for item in parent_type._attribute_names_to_parse.items()
        )

    def _deeply_get_attribute_names(cls) -> dict[str, Callable[[any], any]]:
        """
        Method for getting _attribute_names of own and descendants and subsequent
        caching.
        """

        if '_deeply_attribute_names' in cls.__dict__:
            return cls._deeply_attribute_names

        parent_attribute_names = dict(get_collection_with_reduced_nesting_level_by(
            1,
            (
                parent_type._deeply_get_attribute_names().items()
                for parent_type in cls.__bases__
                if hasattr(parent_type, '_deeply_get_attribute_names')
            )
        ))

        result = parent_attribute_names | cls._attribute_names_to_parse
        cls._deeply_attribute_names = result

        return result

    def _parse_collection_by_attribute_name_from(cls, attributes: dict, attribute_name_to_parse: str) -> tuple:
        """
        Method for getting the final form of a collection by attribute name,
        including attribute elements with this name of all ancestors.
        """

        return (
            tuple(get_collection_with_reduced_nesting_level_by(
                1,
                (
                    getattr(parent_type, attribute_name_to_parse)
                    for parent_type in cls.__bases__
                    if hasattr(parent_type, attribute_name_to_parse)
                )
            ))
            + cls._get_collection_by_attribute_name_from(
                attribute_name_to_parse,
                is_changing=True
            )
        )

    def _get_collection_by_attribute_name_from(
        cls,
        attribute_name_to_parse: str,
        is_changing: bool = False
    ) -> tuple:
        """
        Method for getting all elements of an attribute collection by the input
        name of it from parents, with the ability to change these elements with
        the specified inherited value changer or the default inherent in the
        current class.
        """

        value_getter = cls._deeply_get_attribute_names()[attribute_name_to_parse]

        return tuple(
            (value_getter if value_getter is not None else cls._default_value_transformer)(cls, item)
            for item in cls.__dict__.get(attribute_name_to_parse, tuple())
        )


class CreatingAttributesTransmitterMeta(AttributesTransmitterMeta):
    """
    AttributesTransmitterMeta Metaclass with default transformation of elements
    by creating them if the value is a factory.
    """

    _default_value_transformer = lambda attribute_keeper, original_value: (
        original_value() if isinstance(original_value, CustomArgumentFactory)
        else original_value
    )


class SeparateThreadedLoop(ILoop):
    """
    Proxy implementation of the Loop interface that runs a real implementation
    of that interface on a separate thread. Uses _thread_factory to get the
    desired thread.
    """

    _thread_factory: Callable[[Callable], Thread] = CustomFactory(lambda target: Thread(target=target))

    def __init__(self, loop: ILoop):
        """Gets the implementation of the loop and gets the thread for further work."""

        self._loop = loop
        self._thread = self._thread_factory(loop.run)

    @property
    def thread(self) -> Thread:
        return self._thread

    def run(self) -> None:
        self._thread.start()

    def finish(self) -> None:
        self._loop.finish()
        self._thread.join()


class Loop(ILoop, ABC):
    """Loop class implemented via standard \"while\"."""

    _is_working = False

    def run(self) -> None:
        """Method for starting a while loop with a boolean condition _is_working."""

        self._is_working = True

        while self._is_working:
            self._handle()

    def finish(self) -> None:
        """
        Method for terminating a loop by \"turning off\" the loop condition
        created by the run method.
        """

        self._is_working = False

    @abstractmethod
    def _handle(self) -> None:
        """Method executing inside the \"while\" body and being its representation."""


class HandlerLoop(Loop, ABC):
    """
    Class with the implementation of the "while" body method through a chain of
    actions represented as its special handlers.

    Creates its own handlers through _handlers_factories factories and leaves the
    responsibility for the implementation of receiving them to the heirs.
    """

    _handlers_factories: Iterable[Callable[[Self], 'LoopHandler']]

    def __init__(self):
        self.__handlers = tuple(
            handlers_factory(self)
            for handlers_factory in self._handlers_factories
        )

    @property
    def handlers(self) -> tuple['LoopHandler']:
        return self.__handlers

    def _handle(self) -> None:
        for handler in self.handlers:
            handler.update()


class StrictHandlerLoop(HandlerLoop, metaclass=AttributesTransmitterMeta):
    """
    Descendant of HandlerLoop that implements getting _handlers_factories through
    inheritance and strict indication of them in class fields backed by
    AttributesTransmitterMeta.
    """

    _attribute_names_to_parse = ('_handlers_factories', )


class CustomHandlerLoop(HandlerLoop):
    """
    HandlerLoop is a successor that implements getting _handlers_factories
    through the corresponding input argument.
    """

    def __init__(self, handlers_factories: Iterable[Callable[[HandlerLoop], 'LoopHandler']]):
        self._handlers_factories = handlers_factories
        super().__init__()


class LoopHandler(IUpdatable, ABC):
    """
    Special handler for body variation of HandlerLoop.

    Differs from a regular unit by having a reference to any HandlerLoop.
    Leaves action implementation in update method.
    """

    def __init__(self, loop: HandlerLoop):
        self._loop = loop

    @property
    def loop(self) -> HandlerLoop:
        return self._loop


class UpdaterLoopHandler(LoopHandler):
    """LoopHandler that starts processing of input units."""

    def __init__(self, loop: HandlerLoop, units: Iterable[IUpdatable]):
        self.units = tuple(units)

    def update(self) -> None:
        for unit in self.units:
            unit.update()


class SleepLoopHandler(LoopHandler, ABC):
    """LoopHandler implementing loop delay."""

    def update(self) -> None:
        self._handle_sleep_conditions()

        if self.is_ready_to_sleep():
            self._sleep()

    @abstractmethod
    def is_ready_to_sleep(self) -> bool:
        """Method for determining readiness for the beginning of the delay."""

    @abstractmethod
    def _handle_sleep_conditions(self) -> None:
        """
        Method of processing wait conditions that is not related to the direct
        receipt of waiting readiness.
        """

    @abstractmethod
    def _sleep(self) -> None:
        """Method with loop delay implementation."""


class AlwaysReadyForSleepLoopHandler(SleepLoopHandler, ABC):
    """
    SleepLoopHandler child class with the implementation of a permanent
    completed preparation for sleep.
    """

    def is_ready_to_sleep(self) -> bool:
        return True

    def _handle_sleep_conditions(self) -> None:
        pass


class RollbackSleepLoopHandler(SleepLoopHandler, ABC):
    """
    SleepLoopHandler child class with the ability to roll back sleep conditions
    after the same sleep, if its condition is still active.
    """

    def update(self) -> None:
        super().update()

        if self.is_ready_to_sleep():
            self._sleep_rollback()

    @abstractmethod
    def _sleep_rollback(self) -> None:
        """Method that implements rollback of sleep conditions."""


class TicksSleepLoopHandler(SleepLoopHandler):
    """
    SleepLoopHandler child class delegating the implementation of the wait.

    Uses abstract \"ticks\" as a measure of the degree of wait, matching the
    delegate interface. The immediate delegate is _sleep_function.
    """

    _sleep_function: Callable[[int | float], any]

    def __init__(self, loop: HandlerLoop, ticks_to_sleep: int | float):
        super().__init__(loop)
        self.ticks_to_sleep = ticks_to_sleep

    def _sleep(self) -> None:
        self._sleep_function(self.ticks_to_sleep)


class StandardSleepLoopHandler(TicksSleepLoopHandler, AlwaysReadyForSleepLoopHandler):
    """TicksSleepLoopHandler child class using the standard sleep function."""

    _sleep_function = sleep


class CustomTicksSleepLoopHandler(TicksSleepLoopHandler):
    """TicksSleepLoopHandler child class using the input _sleep_function."""

    def __init__(
        self,
        loop: HandlerLoop,
        ticks_to_sleep: int | float,
        sleep_function: Callable[[int | float], None]
    ):
        self._sleep_function = sleep_function
        super().__init__(loop, ticks_to_sleep)


class TickerSleepLoopHandler(TicksSleepLoopHandler, RollbackSleepLoopHandler, ABC):
    """
    TicksSleepLoopHandler child class that implements a condition to wait after
    a certain specified time.

    Strictly binds \"ticks\" to the amount of updating its state, through which
    it is implemented. Leaves the possibility for descendants to change the
    quantitative assignment of updating their state to a \"tick\".
    """

    _real_ticks_to_sleep: int | float = 0
    _tick: int | float = 1

    @property
    def real_ticks_to_sleep(self) -> int | float:
        return self._real_ticks_to_sleep

    @property
    def ticks_to_sleep(self) -> int | float:
        return self.__ticks_to_sleep

    @ticks_to_sleep.setter
    def ticks_to_sleep(self, ticks_to_sleep: int | float) -> None:
        if self._real_ticks_to_sleep > ticks_to_sleep:
            self._real_ticks_to_sleep = ticks_to_sleep

        self.__ticks_to_sleep = ticks_to_sleep

    def is_ready_to_sleep(self) -> bool:
        return self._real_ticks_to_sleep <= 0

    def _handle_sleep_conditions(self) -> None:
        self._real_ticks_to_sleep -= self._tick

    def _sleep_rollback(self) -> None:
        self._real_ticks_to_sleep = self.ticks_to_sleep


class CustomTickerSleepLoopHandler(CustomTicksSleepLoopHandler, TickerSleepLoopHandler):
    """
    Class that is an addition of TickerSleepLoopHandler to a custom
    implementation of its parent for a custom version of the
    TickerSleepLoopHandler itself.
    """


class INumberRounder(ABC):
    """Interface that allows you to somehow round the entered number."""

    @abstractmethod
    def __call__(self, number: any) -> any:
        """Gateway call method to start rounding."""


class NumberRounder(INumberRounder, ABC):
    """
    Base class for INumberRounder interface.

    Implements delegation of the main rounding method to a protected
    corresponding method.
    """

    def __call__(self, number: any) -> any:
        return self._round(number)

    @abstractmethod
    def _round(self, number: int | float) -> float:
        """Internal rounding method that implements rounding of a number."""


class FastNumberRounder(NumberRounder):
    """
    NumberRounder child class that implements rounding through the standard
    floor function.
    """

    def _round(self, number: int | float) -> float:
        return floor(number)


class AccurateNumberRounder(NumberRounder):
    """
    NumberRounder child class that implements rounding through string conversion
    and subsequent determination of the rounding side.
    """

    def _round(self, number: int | float) -> float:
        number_after_point = int(str(float(number)).split('.')[1][0])

        if number_after_point >= 5:
            return int(number) + copysign(1, number)
        else:
            return int(number)


class ProxyRounder(NumberRounder):
    """
    Proxy NumberRounder delegating the rounding to real NumberRounder.
    Gets a delegate during initialization.
    """

    def __init__(self, rounder: NumberRounder):
        self.rounder = rounder

    def _round(self, number: int | float) -> float:
        return self.rounder(number)


class ShiftNumberRounder(ProxyRounder):
    """ProxyRounder child class that implements rounding to a certain degree."""

    def __init__(self, rounder: NumberRounder, comma_shift: int):
        super().__init__(rounder)
        self.comma_shift = comma_shift

    def _round(self, number: int | float) -> float:
        return self.__move_point_in_number(
            super()._round(
                self.__move_point_in_number(number, self.comma_shift)
            ),
            -self.comma_shift
        )

    def __move_point_in_number(self, number: int | float, shift: int) -> float:
        """Method for moving a dot in the input numebr."""

        letters_of_number = list(str(float(number)))
        point_index = letters_of_number.index('.')
        letters_of_number.pop(point_index)

        point_index += shift

        if point_index > len(letters_of_number):
            letters_of_number.extend(
                ('0' for _ in range(point_index - len(letters_of_number)))
            )
        elif point_index < 0:
            point_index = 0

        letters_of_number.insert(point_index, '.')

        return float(''.join(letters_of_number))


@dataclass
class Report:
    """
    Structure for storing and passing data about the state of something before
    further processing.
    """

    sign: bool
    message: str | None = None
    error: Exception | None = None

    def __bool__(self) -> bool:
        return self.sign

    @classmethod
    def create_error_report(cls, error: Exception) -> Self:
        """Method for quickly creating a report from a ready-made error."""

        return cls(
            False,
            error=error
        )


class ReportHandler(ABC):
    """Base class of a report handler."""

    @abstractmethod
    def __call__(self, report: Report) -> None:
        pass

    def is_supported_report(self, report: Report) -> bool:
        return True


class BadReportHandler(ReportHandler):
    """
    ReportHandler child class that strictly reacts to negative reports by
    raising an error that is present in the report or made according to the
    report model.

    During initialization, it receives the default data filled into the error
    that has arisen if strictly specified in the "bad" report is left out.
    """

    def __init__(
        self,
        default_error_type: type,
        default_error_message: str = ''
    ):
        self.default_error_type = default_error_type
        self.default_error_message = default_error_message

    def __call__(self, report: Report) -> None:
        if report.error:
            raise report.error

        raise self.default_error_type(
            report.message if report.message else self.default_error_message
        )

    def is_supported_report(self, report: Report) -> bool:
        return not report.sign

    @classmethod
    def create_from_error(cls, error: Exception) -> Self:
        return cls(type(error), str(error))


class ReportAnalyzer:
    """Action chain class from report handlers for, respectively, reports."""

    def __init__(self, report_handlers: Iterable[ReportHandler]):
        self.report_handlers = frozenset(report_handlers)

    def __call__(self, report: Report) -> None:
        for report_handler in self.report_handlers:
            if report_handler.is_supported_report(report):
                report_handler(report)


class IReporter(ABC):
    """Interface for creating a report based on a set of objects."""

    @abstractmethod
    def create_report_of(self, objects: Iterable) -> Report:
        """Method of corresponding report receipt on objects."""


class ProxyReporter(IReporter):
    """
    Proxy reporter to structure multiple reports from multiple reporters.

    Works like "and" or "or" for newly created reports. The is_unanimous
    attribute is responsible for the selected mod. Enabled is_unanimous is
    responsible for the "and" mod, Disabled, respectively, for "or" mod.
    """

    def __init__(self, reporters: Iterable[IReporter], is_unanimous: bool = True):
        self.is_unanimous = is_unanimous
        self._reporters = tuple(reporters)

    @property
    def reporters(self) -> tuple[IReporter]:
        return self._reporters

    def create_report_of(self, objects: Iterable) -> Report:
        for reporter in self._reporters:
            report = reporter.create_report_of(objects)

            if (
                report.sign and not self.is_unanimous
                or not report.sign and self.is_unanimous
            ):
                return report
        else:
            return report


class CallableProxyReporter(ProxyReporter):
    """ProxyReporter with the ability to call it as a function."""

    def __call__(self, objects: Iterable) -> Report:
        return self.create_report_of(objects)


class FocusedProxyReporter(ProxyReporter):
    """ProxyReporter with an interface calling to report on a single object."""

    def __call__(self, object_: object) -> Report:
        return self.create_report_of((objects, ))


class TypeReporter(StylizedMixin, IReporter):
    """
    Reporter class checking the correctness of the types of the input objects to
    its stored ones.

    Has two strictness modes for matching stored tops: to all or only to one.
    Selects a mod by the corresponding input is_all_types_needed attribute.
    """

    _repr_fields = Field(
        'supported_types',
        value_transformer=lambda types: str(tuple(type.__name__ for type in types)).replace('\'', '')
    ),

    def __init__(self, supported_types: Iterable[type], is_all_types_needed: bool = False):
        self.is_all_types_needed = is_all_types_needed
        self.supported_types = supported_types

    @property
    def supported_types(self) -> tuple[type]:
        return self.__supported_types

    @supported_types.setter
    def supported_types(self, new_types: Iterable[type]) -> None:
        self.__supported_types = tuple(new_types)
        self._update_report_message()

    def create_report_of(self, objects: Iterable) -> Report:
        return Report(
            (all if self.is_all_types_needed else any)(
                isinstance(object_, supported_type)
                for object_ in objects
                for supported_type in self.supported_types
            ),
            self._report_message
        )

    def _update_report_message(self) -> None:
        """Report message pre-creation method."""

        base_template = "object type must be {type_names}"
        type_names = tuple(map(
            lambda type_: type_.__name__, self.supported_types
        ))

        self._report_message = "object type must be {}".format(
            "{first_type_names} {set_type_determinant_word} {last_type_name}".format(
                first_type_names=', '.join(type_names[:-1]),
                last_type_name=type_names[-1],
                set_type_determinant_word='and' if self.is_all_types_needed else 'or'
            ) if len(self.supported_types) > 1 else ', '.join(type_names)
        )


class TypeReporterKeeperMeta(AttributesTransmitterMeta):
    """
    Metaclass for native creation of type_reporter from the fields of the used
    class.

    Used classes must have the _supported_types attribute.
    """

    _attribute_names_to_parse = '_suported_types',
    _type_reporter_factory: Callable[[tuple[type]], TypeReporter] = TypeReporter

    def __new__(cls, class_name: str, super_classes: tuple, attributes: dict):
        isinstance_type = super().__new__(cls, class_name, super_classes, attributes)
        isinstance_type._type_reporter = TypeReporter(isinstance_type._suported_types)

        return isinstance_type

    @property
    def type_reporter(cls) -> TypeReporter:
        return cls._type_reporter


class StrictToStateMixin(ABC):
    """
    Mixin class that implements handling of object state reports.

    Child Class must have _state_report_analyzer attribute to perform handling.
    """

    _state_report_analyzer: ReportAnalyzer

    @abstractmethod
    def _is_correct(self) -> Report:
        """Method for creating object state reports."""

    def _check_state_errors(self) -> None:
        """Method that starts handling the state of current object."""

        self._state_report_analyzer(self._is_correct())


class Divider(ABC):
    """Base class that allows you to share any data."""

    _report_analyzer = ReportAnalyzer((BadReportHandler(UnableToDivideError), ))

    def __call__(self, data: any) -> None:
        self._report_analyzer(self.is_possible_to_divide(data))
        return self._divide(data)

    def is_possible_to_divide(self, data: any) -> Report:
        return Report(True)

    @abstractmethod
    def _divide(self, data: any) -> None:
        """Data division method."""


class ComparisonResult(IntEnum):
    """Flags to describe comparison results."""

    less = -1
    equals = 0
    more = 1


def compare(main: any, relatival: any) -> ComparisonResult:
    """
    Comparer function that returns the result of a comparison with the
    ComparisonResult flag. Compares relatival from main
    """

    if main > relatival:
        return ComparisonResult.more
    elif main < relatival:
        return ComparisonResult.less
    else:
        return ComparisonResult.equals


@dataclass(frozen=True)
class RGBAColor:
    """
    Structure for storing color data in RGBA format and maintaining the data in
    the correct form.
    """

    red: int = 0
    green: int = 0
    blue: int = 0
    alpha_channel: float = 1.

    def __post_init__(self) -> None:
        if any(
            not (0 <= color_coordinate <= 255)
            for color_coordinate in (self.red, self.green, self.blue)
        ):
            raise ColorCoordinateError(
                f"Color coordinate must be between 0 and 255"
            )
        elif not 0 <= self.alpha_channel <= 1:
            raise AlphaChannelError("Alpha channel must be between 0 and 1")

    def __iter__(self) -> iter:
        return iter((self.red, self.green, self.blue, self.alpha_channel))


def like_object(func: Callable) -> Callable:
    """Decorator passing a link of the input function to it."""

    @wraps(func)
    def wrapper(*args, **kwargs) -> any:
        return func(func, *args, **kwargs)

    return wrapper


class Timer(StylizedMixin):
    """Class that implements a countdown."""

    _repr_fields = (
        Field('period', value_transformer=lambda value: f"{value} second{'s' if value > 1 else ''}"),
        Field('end_time', value_transformer=ctime)
    )

    def __init__(self, seconds_of_period: int):
        self.period = seconds_of_period
        self._end_time = 0
        self.start()

    @property
    def end_time(self) -> float:
        return self._end_time

    def is_time_over(self) -> bool:
        return self._end_time <= time()

    def start(self):
        if not self.is_time_over():
            raise TimerError(f"Timer {self} has already started")

        self._end_time = time() + self.period


class Comparable(Protocol):
    """Protocol for annotating comparable types."""

    @abstractmethod
    def __eq__(self, other: int | float | Self) -> bool:
        pass

    @abstractmethod
    def __ne__(self, other: int | float | Self) -> bool:
        pass

    @abstractmethod
    def __lt__(self, other: int | float | Self) -> bool:
        pass

    @abstractmethod
    def __le__(self, other: int | float | Self) -> bool:
        pass

    @abstractmethod
    def __gt__(self, other: int | float | Self) -> bool:
        pass

    @abstractmethod
    def __ge__(self, other: int | float | Self) -> bool:
        pass


class Diapason(StylizedMixin):
    """Class to simulate a range."""

    _repr_fields = Field(
        value_getter=lambda diapason, _: (diapason.start, diapason.end),
        formatter=lambda value, _: ' ~ '.join(map(str, value))
    ),

    def __init__(self, first: Comparable, second: Comparable = 0, is_end_inclusive: bool = False):
        self.is_end_inclusive = is_end_inclusive
        self.update_by(first, second)

    def __contains__(self, something: Comparable) -> bool:
        return self.is_having(something)

    @property
    def start(self) -> Comparable:
        return self._start

    @property
    def end(self) -> Comparable:
        return self._end

    def update_by(self, first: Comparable, second: Comparable):
        self._start, self._end = sorted((first, second))

    def is_having(self, something: Comparable) -> bool:
        return (
            something >= self._start
            and (
                something < self._end
                or (something == self._end if self.is_end_inclusive else False)
            )
        )



from dataclasses import dataclass
from typing import Iterable

from sim32.basic_render_resources import *
from sim32.geometry import Vector
from sim32.tools import RGBAColor


@dataclass
class PygamePolygon(Polygon):
    """Dataclass containing data for pygame polygon drawing function."""

    border_width: int | float = 0


@dataclass
class PygameLine(Line):
    """
    Dataclass containing data for pygame line and aaline drawing functions.

    The value of is_smooth attribute defines the annotation on the using function:
    True - aaline, False - line.
    """

    border_width: int | float = 1
    is_smooth: bool = False


@dataclass
class PygameLines(ColorRenderResource):
    """
    Dataclass containing data for pygame lines and aalines drawing functions.

    The value of is_smooth attribute defines the annotation on the using function:
    True - aalines, False - lines.
    """

    is_closed: bool
    points: Iterable[Vector]
    border_width: int | float = 1
    is_smooth: bool = False


@dataclass
class PygameCircle(Circle):
    """Dataclass containing data for pygame circle drawing function."""

    border_width: int | float = 0


@dataclass
class PygameRectangle(Rectangle):
    """Dataclass containing data for pygame rect drawing function."""

    border_width: int | float = 0


@dataclass
class PygameEllipse(Rectangle):
    """Dataclass containing data for pygame ellipse drawing function."""

    border_width: int | float = 0


@dataclass
class PygameArc(Rectangle):
    """Dataclass containing data for pygame arc drawing function."""

    start_angle: int | float
    stop_angle: int | float
    border_width: int | float = 1



from abc import ABC
from dataclasses import dataclass
from typing import Callable, Iterable, Self

from beautiful_repr import StylizedMixin, Field

from sim32.core import MultitaskingUnit
from sim32.geometry import Vector
from sim32.renders import ResourcePack
from sim32.interfaces import IAvatar, IAvatarKeeper
from sim32.errors.avatar_errors import AnimationAlreadyFinishedError


class Avatar(IAvatar, ABC):
    """Base avatar with domain object acquisition implementation."""

    def __init__(self, domain: IAvatarKeeper):
        self._domain = domain

    @property
    def domain(self) -> IAvatarKeeper:
        return self._domain


class SingleResourcePackAvatar(Avatar, ABC):
    """Avatar class using only one resource pack."""

    _main_resource_pack: ResourcePack

    @property
    def render_resource_packs(self) -> tuple[ResourcePack]:
        return (self._main_resource_pack, )

    def update(self) -> None:
        self._main_resource_pack.point = self.domain.position


class ResourceAvatar(SingleResourcePackAvatar, StylizedMixin, ABC):
    """
    Avatar class operating on only one render resource.

    To create a resource and its pack uses the appropriate factories from
    _resource_factory and _resource_pack_factory attributes.
    """

    _repr_fields = (Field('resource'), )
    _resource_factory: Callable[[Self], any]
    _resource_pack_factory: Callable[[any, Vector], ResourcePack] = ResourcePack

    def __init__(self, domain: IAvatarKeeper):
        super().__init__(domain)
        self._main_resource_pack = self._resource_pack_factory(
            self._resource_factory(self),
            self.domain.position
        )

    @property
    def render_resource(self) -> any:
        """Property to get one available render resource."""

        return self._main_resource_pack.resource


class PrimitiveAvatar(ResourceAvatar):
    """Avatar class that wraps the position of a domain and an input render resource."""

    def __init__(self, domain: IAvatarKeeper, resource: any):
        self._resource_factory = lambda _: resource
        super().__init__(domain)

    @property
    def render_resource(self) -> any:
         return ResourceAvatar.render_resource.fget(self)

    @render_resource.setter
    def render_resource(self, render_resource: any) -> None:
        self._main_resource_pack.resource = render_resource


@dataclass
class Sprite:
    """Dataclass pack render resource processed as a sprite."""

    resource: any
    max_stay_ticks: int

    def __post_init__(self):
        self.real_stay_ticks = self.max_stay_ticks


class Animation(SingleResourcePackAvatar, ABC):
    """
    Avatar class implemented as an animation from sprites.

    Leaves the sprite storage implementation to childs. Throws an error when
    playing all sprites.
    """

    _sprites: tuple[Sprite]
    _current_sprite_index: int = 0

    def __init__(self, domain: IAvatarKeeper):
        super().__init__(domain)
        self._update_main_resource_pack()

    def update(self) -> None:
        super().update()

        if self.is_finished():
            self._handle_finish()

        self._update_main_resource_pack()

        self._active_sprite.real_stay_ticks -= 1

        if self._active_sprite.real_stay_ticks <= 0:
            self._current_sprite_index += 1

    def is_finished(self) -> bool:
        return (
            self._current_sprite_index >= len(self._sprites) - 1
            and self._active_sprite.real_stay_ticks <= 0
        )

    @property
    def _active_sprite(self) -> Sprite:
        """Property for the current animation sprite."""

        return self._sprites[self._current_sprite_index]

    def _update_main_resource_pack(self) -> None:
        """Method for updating the main resource pack."""

        self._main_resource_pack = ResourcePack(
            self._sprites[self.__current_sprite_index].resource,
            self.domain.position
        )

    def _handle_finish(self) -> None:
        """
        Method for handling the end of sprites for assignment to the main
        main_resource_pack.
        """

        raise AnimationAlreadyFinishedError(f"Animation {self} already finished")


class CustomAnimation(Animation):
    """Animation class with input sprites."""

    def __init__(self, domain: IAvatarKeeper, sprites: Iterable[Sprite]):
        super().__init__(domain)
        self._sprites = tuple(sprites)


class EndlessAnimation(Animation, ABC):
    """Animation class for looping sprites."""

    def _handle_finish(self) -> None:
        self._current_sprite_index = 0

        for sprite in self._sprites:
            sprite.real_stay_ticks = sprite.max_stay_ticks


class CustomEndlessAnimation(EndlessAnimation, CustomAnimation):
    """Endless Animation class with input sprites."""


class AnimationAvatar(Avatar, ABC):
    """Avatar class delegating responsibilities to animations."""

    _default_animation_factory: Callable[[IAvatarKeeper], EndlessAnimation]

    def __init__(self, domain: IAvatarKeeper):
        super().__init__(domain)
        self._current_animation = self._default_animation = self._default_animation_factory(domain)

    @property
    def render_resource_packs(self) -> tuple[ResourcePack]:
        return self._current_animation.render_resource_packs

    def update(self) -> None:
        self._current_animation.update()


class TopicAnimationAvatar(AnimationAvatar, ABC):
    """Animation Avatar class that implements selection of animations by topic."""

    _animation_factory_by_topic: dict[str, Callable[[IAvatarKeeper], EndlessAnimation]]

    def __init__(self, domain: IAvatarKeeper):
        super().__init__(domain)

        self._animation_by_topic = {
            topic: animation_factory(domain)
            for topic, animation_factory in self._animation_factory_by_topic.items()
        }

    def update(self) -> None:
        if (
            self._default_animation is not self._current_animation
            and self._current_animation.is_finished()
        ):
            self._current_animation = self._default_animation

        super().update()

    def activate_animation_by_topic(self, topic: str) -> None:
        """Animation selection method by topic."""

        self._current_animation = self._animation_by_topic[topic]


class CustomTopicAnimationAvatar(TopicAnimationAvatar):
    """Topic Animation Avatar class with input animation factories and topics."""

    def __init__(
        self,
        domain: IAvatarKeeper,
        animation_factory_by_topic: dict[str, Callable[[IAvatarKeeper], EndlessAnimation]]
    ):
        self._animation_factory_by_topic = animation_factory_by_topic
        super().__init__(domain)


class ProcessAnimationAvatar(AnimationAvatar):
    """
    Animation Avatar class that implements the choice of animations for the
    processes running in the domain.
    """

    _animation_factory_by_process_type: dict[type, Callable[[MultitaskingUnit], EndlessAnimation]]

    def __init__(self, domain: IAvatarKeeper):
        super().__init__(domain)

        self.__domains_previous_processes = domain.processes
        self._animation_by_process_type = {
            process_type: animation_factory(domain)
            for process_type, animation_factory in self._animation_factory_by_process_type.items()
        }

    def update(self) -> None:
        for process in self.domain.processes - self.__domains_previous_processes:
            if type(process) in self._animation_by_process_type:
                self._current_animation = self._animation_by_process_type[type(process)]
                break

        self.__domains_previous_processes = self.domain.processes

        super().update()
